r"""
    Byte-streams and snapshots used in DAG-CBOR decoding, keeping track of latest and previous read byte chunks for error reporting purposes.
"""

from __future__ import annotations # See https://peps.python.org/pep-0563/

from io import BufferedIOBase, BytesIO
from typing import Optional

class StreamSnapshot:
    r""" A snapshot of the current state of a byte-stream being decoded. """

    _bs: bytes
    _pos: int

    def __new__(cls, latest_read: bytes, next_read_start: int) -> "StreamSnapshot":
        instance = object.__new__(cls)
        instance._bs = latest_read
        instance._pos = next_read_start
        return instance

    @property
    def latest_read(self) -> bytes:
        r""" The latest byte chunk read from the stream. """
        return self._bs

    @property
    def latest_read_size(self) -> int:
        r""" Size of the latest byte chunk read from the stream. """
        return len(self._bs)

    @property
    def latest_read_start(self) -> int:
        r""" Start position in the stream for the latest byte chunk read. """
        return self._pos-len(self._bs)

    @property
    def num_bytes_read(self) -> int:
        r""" Total number of bytes read so far from the stream. """
        return self._pos


class Stream:
    r"""
        Container for the byte-stream being decoded, offering additional book-keeping functionality used to produce detailed error messages.
    """

    _buf: BufferedIOBase
    _bs: bytes
    _pos: int
    _prev_bs: bytes
    _prev_pos: int

    def __new__(cls, buffer: Optional[BufferedIOBase] = None, init_bytes_read: bytes = bytes()) -> "Stream":
        if buffer is None:
            buffer = BytesIO(bytes())
        instance = object.__new__(cls)
        instance._buf = buffer
        instance._bs = init_bytes_read
        instance._pos = len(init_bytes_read)
        instance._prev_bs = bytes()
        instance._prev_pos = 0
        return instance

    @property
    def curr_snapshot(self) -> "StreamSnapshot":
        r""" A snapshot of the current state of the stream. """
        return StreamSnapshot(self._bs, self._pos)

    @property
    def prev_snapshot(self) -> "StreamSnapshot":
        r""" A snapshot of the state of the stream immediately before the latest non-extending read. """
        return StreamSnapshot(self._prev_bs, self._prev_pos)

    def read(self, num_bytes: Optional[int] = None, *, extend: bool = False) -> bytes:
        r"""
            Read the given number of bytes from the stream. If :obj:`None`, reads all remaining bytes.
            If ``extend`` is set to :obj:`True`, the current stream snapshot (see :attr:`Stream.curr_snapshot`) is extended with the bytes just read,
            and the previous stream snapshot (see :attr:`Stream.prev_snapshot`) is kept.
            Otherwise, the previous snapshot is replaced with the current snaptshot, and a new current snapshot is created with the bytes just read.
        """
        bs = self._buf.read(num_bytes)
        if extend:
            self._bs += bs
            self._pos += len(bs)
        else:
            self._prev_bs = self._bs
            self._prev_pos = self._pos
            self._bs = bs
            self._pos += len(bs)
        return bs



"""
    Encoding functions for DAG-CBOR codec.
"""

from __future__ import annotations # See https://peps.python.org/pep-0563/

from io import BufferedIOBase, BytesIO
import math
import struct
from typing import Any, Dict, List, Optional, overload, Union
import unicodedata

from typing_extensions import Literal, TypedDict
from typing_validation import validate

from multiformats import varint, multicodec, CID

from ..ipld import IPLDKind, IPLDObjPath
from .err import CBOREncodingError, DAGCBOREncodingError

__all__ = ("CBOREncodingError", "DAGCBOREncodingError")

_dag_cbor_multicodec = multicodec.get("dag-cbor")
_dag_cbor_code: int = _dag_cbor_multicodec.code
_dag_cbor_code_bytes: bytes = varint.encode(_dag_cbor_code)
_dag_cbor_code_nbytes: int = len(_dag_cbor_code_bytes)

def check_key_compliance(value: Dict[str, Any]) -> None:
    """
        Enforces DAG-CBOR compliance for keys in a mapping.
    """
    validate(value, Dict[str, Any])
    _check_key_compliance(value)

def canonical_order_dict(value: Dict[str, Any]) -> Dict[str, Any]:
    """
        Returns a dictionary with canonically ordered keys, according to the DAG-CBOR specification.
        Specifically, keys are sorted increasingly first by length and then by the lexicographic ordering of the corresponding UTF-8 bytestrings.
    """
    validate(value, Dict[str, Any])
    _check_key_compliance(value)
    # sort keys canonically
    return _canonical_order_dict(value)


@overload
def encode(data: IPLDKind, stream: None = None, *,
           include_multicodec: bool = False,
           normalize_strings: Optional[Literal["NFC", "NFKC", "NFD", "NFKD"]] = None
          ) -> bytes:
    ... # pragma: no cover

@overload
def encode(data: IPLDKind, stream: BufferedIOBase, *,
           include_multicodec: bool = False,
           normalize_strings: Optional[Literal["NFC", "NFKC", "NFD", "NFKD"]] = None
          ) -> int:
    ... # pragma: no cover

def encode(data: IPLDKind, stream: Optional[BufferedIOBase] = None, *,
           include_multicodec: bool = False,
           normalize_strings: Optional[Literal["NFC", "NFKC", "NFD", "NFKD"]] = None
          ) -> Union[bytes, int]:
    r"""
        Encodes the given data with the DAG-CBOR codec.

        By default, the encoded data is written to an internal stream and the bytes are returned at the end (as a `bytes` object).

        .. code-block:: python

            def encode(data: IPLDKind, stream: None = None) -> bytes:
                ...

        Example usage:

        >>> dag_cbor.encode({'a': 12, 'b': 'hello!'})
        b'\xa2aa\x0cabfhello!'

        If a ``stream`` is given, the encoded data is written to the stream and the number of bytes written is returned:

        .. code-block:: python

            def encode(data: IPLDKind, stream: BufferedIOBase) -> int:
                ...

        Example usage with a stream:

        >>> from io import BytesIO
        >>> stream = BytesIO()
        >>> dag_cbor.encode({'a': 12, 'b': 'hello!'}, stream=stream)
        13
        >>> stream.getvalue()
        b'\xa2aa\x0cabfhello!'

        :param data: the DAG data to be encoded
        :param stream: an optional stream into which the encoded data should be written
        :param include_multicodec: if :obj:`True`, the encoded data is prefixed by the multicodec code for ``'dag-cbor'``
                                   (see `multicodec.wrap <https://multiformats.readthedocs.io/en/latest/api/multiformats.multicodec.html#wrap>`_).
        :param normalize_strings: whether strings should be normalised prior to encoding

        :raises CBOREncodingError: if an :obj:`int` outside of ``range(-2**64, 2**64)`` is encountered
        :raises DAGCBOREncodingError: if a value of type other than :obj:`None`, :obj:`bool`, :obj:`int`, :obj:`float`, :obj:`str`,
                                                      :obj:`bytes`, :obj:`list`, :obj:`dict`, or :class:`~multiformats.cid.CID` is encountered
        :raises DAGCBOREncodingError: if attempting to encode the special :obj:`float` values ``NaN``, ``Infinity`` and ``-Infinity``
        :raises DAGCBOREncodingError: if a key of a dictionary is not a string

    """
    validate(stream, Optional[BufferedIOBase])
    validate(include_multicodec, bool)
    options: _EncodeOptions = {}
    if normalize_strings is not None:
        validate(normalize_strings, Literal["NFC", "NFKC", "NFD", "NFKD"])
        options["normalize_strings"] = normalize_strings
    path = IPLDObjPath()
    if stream is None:
        internal_stream = BytesIO()
        if include_multicodec:
            internal_stream.write(_dag_cbor_code_bytes)
        _encode(internal_stream, data, path, options)
        return internal_stream.getvalue()
    num_bytes = 0
    if include_multicodec:
        stream.write(_dag_cbor_code_bytes)
        num_bytes += _dag_cbor_code_nbytes
    num_bytes += _encode(stream, data, path, options)
    return num_bytes

class _EncodeOptions(TypedDict, total=False):
    r""" Options passed around to encoding sub-routines. """

    normalize_strings: Literal["NFC", "NFKC", "NFD", "NFKD"]
    r""" Optional Unicode normalization to be performed on UTF-8 strings prior to byte encoding. """

def _encode(stream: BufferedIOBase, value: IPLDKind, path: IPLDObjPath, options: _EncodeOptions) -> int:
    # pylint: disable = too-many-return-statements, too-many-branches
    if isinstance(value, bool): # must go before int check
        # major type 0x7 (additional info 20 and 21)
        return _encode_bool(stream, value, path, options)
    if isinstance(value, int):
        # major types 0x0 and 0x1
        return _encode_int(stream, value, path, options)
    if isinstance(value, bytes):
        # major type 0x2
        return _encode_bytes(stream, value, path, options)
    if isinstance(value, str):
        # major type 0x3
        return _encode_str(stream, value, path, options)
    if isinstance(value, list):
        # major type 0x4
        return _encode_list(stream, value, path, options)
    if isinstance(value, dict):
        # major type 0x5
        return _encode_dict(stream, value, path, options)
    if isinstance(value, CID):
        # major type 0x6
        return _encode_cid(stream, value, path, options)
    if value is None:
        # major type 0x7 (additional info 22)
        return _encode_none(stream, value, path, options)
    if isinstance(value, float):
        # major type 0x7 (additional info 27)
        return _encode_float(stream, value, path, options)
    err = f"Error encoding value at {path}: value is not of IPLD kind (found type {type(value)})."
    raise DAGCBOREncodingError(err)

def _encode_head(stream: BufferedIOBase, major_type: int, arg: int) -> int:
    if arg < 24:
        # argument value as additional info in leading byte
        head = struct.pack(">B", (major_type<<5)|arg)
    elif arg <= 255:
        # leading byte + 1 byte argument value (additional info = 24)
        head = struct.pack(">BB", (major_type<<5)|24, arg)
    elif arg <= 65535:
        # leading byte + 2 bytes argument value (additional info = 25)
        head = struct.pack(">BH", (major_type<<5)|25, arg)
    elif arg <= 4294967295:
        # leading byte + 4 bytes argument value (additional info = 26)
        head = struct.pack(">BL", (major_type<<5)|26, arg)
    else:
        # leading byte + 8 bytes argument value (additional info = 27)
        head = struct.pack(">BQ", (major_type<<5)|27, arg)
    stream.write(head)
    return len(head)

def _encode_int(stream: BufferedIOBase, value: int, path: IPLDObjPath, options: _EncodeOptions) -> int:
    if value >= 18446744073709551616:
        # unsigned int must be < 2**64
        err = f"Error encoding integer value at {path}: Unsigned integer out of range."
        raise CBOREncodingError(err)
    if value < -18446744073709551616:
        # negative int must be >= -2**64
        err = f"Error encoding integer value at {path}: Negative integer out of range."
        raise CBOREncodingError(err)
    if value >= 0:
        # unsigned int
        return _encode_head(stream, 0x0, value)
    # negative int
    return _encode_head(stream, 0x1, -1-value)

def _encode_bytes(stream: BufferedIOBase, value: bytes, path: IPLDObjPath, options: _EncodeOptions) -> int:
    num_head_bytes = _encode_head(stream, 0x2, len(value))
    stream.write(value)
    return num_head_bytes+len(value)

def _encode_str(stream: BufferedIOBase, value: str, path: IPLDObjPath, options: _EncodeOptions) -> int:
    if "normalize_strings" in options:
        value = unicodedata.normalize(options["normalize_strings"], value)
    utf8_value: bytes = value.encode("utf-8", errors="strict")
    num_head_bytes = _encode_head(stream, 0x3, len(utf8_value))
    stream.write(utf8_value)
    return num_head_bytes+len(utf8_value)

def _encode_list(stream: BufferedIOBase, value: List[Any], path: IPLDObjPath, options: _EncodeOptions) -> int:
    num_bytes_written = _encode_head(stream, 0x4, len(value))
    for idx, item in enumerate(value):
        num_bytes_written += _encode(stream, item, path/idx, options)
    return num_bytes_written

def _encode_dict(stream: BufferedIOBase, value: Dict[str, Any], path: IPLDObjPath, options: _EncodeOptions) -> int:
    _check_key_compliance(value, path)
    if "normalize_strings" in options:
        nf = options["normalize_strings"]
        value = {unicodedata.normalize(nf, k): v for k, v in value.items()}
    utf8key_val_pairs = [(k, k.encode("utf-8", errors="strict"), v)
                         for k, v in value.items()]
    # 1. sort keys canonically:
    sorted_utf8key_val_pairs = sorted(utf8key_val_pairs, key=lambda i: (len(i[1]), i[1]))
    # 2. encode key-value pairs (keys already utf-8 encoded):
    num_bytes_written = _encode_head(stream, 0x5, len(value))
    for k, utf8k, v in sorted_utf8key_val_pairs:
        num_bytes_written += _encode_head(stream, 0x3, len(utf8k))
        stream.write(utf8k)
        num_bytes_written += len(utf8k)
        num_bytes_written += _encode(stream, v, path/k, options)
    return num_bytes_written

def _encode_cid(stream: BufferedIOBase, value: CID, path: IPLDObjPath, options: _EncodeOptions) -> int:
    num_bytes_written = _encode_head(stream, 0x6, 42)
    num_bytes_written += _encode_bytes(stream, b"\0" + bytes(value), path, options)
    return num_bytes_written

def _encode_bool(stream: BufferedIOBase, value: bool, path: IPLDObjPath, options: _EncodeOptions) -> int:
    return _encode_head(stream, 0x7, 21 if value else 20)

def _encode_none(stream: BufferedIOBase, value: None, path: IPLDObjPath, options: _EncodeOptions) -> int:
    return _encode_head(stream, 0x7, 22)

def _encode_float(stream: BufferedIOBase, value: float, path: IPLDObjPath, options: _EncodeOptions) -> int:
    if math.isnan(value):
        err = f"Error encoding float value at {path}: NaN is not allowed."
        raise DAGCBOREncodingError(err)
    if math.isinf(value):
        s = "" if value > 0 else "-"
        err = f"Error encoding float value at {path}: {s}Infinity is not allowed."
        raise DAGCBOREncodingError(err)
    # special head, with double encoding for 4B argument value
    head = struct.pack(">Bd", (0x7<<5)|27, value)
    stream.write(head)
    return len(head)

def _check_key_compliance(value: Dict[str, Any], path: Optional[IPLDObjPath] = None) -> None:
    """ Check keys for DAG-CBOR compliance. """
    for idx, k in enumerate(value.keys()):
        if not isinstance(k, str):
            err = "" if path is None else f"Error encoding value of map kind at {path}: "
            err += f"key for key-value pair at position {idx} is not a string."
            raise DAGCBOREncodingError(err)

def _canonical_order_dict(value: Dict[str, Any]) -> Dict[str, Any]:
    utf8key_key_val_pairs = [(k.encode("utf-8", errors="strict"), k, v) for k, v in value.items()]
    sorted_utf8key_key_val_pairs = sorted(utf8key_key_val_pairs, key=lambda i: (len(i[0]), i[0]))
    return {k: v for _, k, v in sorted_utf8key_key_val_pairs}



# pylint: disable = all
# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

from __future__ import annotations

import inspect
import json
import os
import sys
sys.path.insert(0, os.path.abspath('..'))
import sphinx_rtd_theme # type: ignore


# -- Project information -----------------------------------------------------

project = 'dag-cbor'
copyright = '2023, Hashberg'
author = 'Hashberg'


# The version info for the project you"re documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = "0.3.0"
# The short X.Y version.
version = "0.3.0"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.mathjax',
    'sphinx.ext.intersphinx',
    'sphinx_rtd_theme',
    'sphinx.ext.viewcode',
    'autodoc_typehints',
]

nitpicky = True                  # warn about every broken reference
add_function_parentheses = False # no parentheses after function names
add_module_names = False         # no module names at start of classes
set_type_checking_flag = False   # setting to True creates issues when dealing with circular dependencies introduced for static typechecking
autodoc_typehints = "none"       # don't document type hints, extension 'autodoc_typehints' will take care of it

with open("autodoc-type-aliases.json", "r") as f:
    autodoc_type_aliases = json.load(f) # load type aliases generated by make-api.py

intersphinx_cache_limit = -1
intersphinx_timeout = 20
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'multiformats': ('http://multiformats.readthedocs.io/en/latest', None)
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "test"
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']



r"""
    Detailed messages for all possible DAG-CBOR decoding errors.
"""

from __future__ import annotations # See https://peps.python.org/pep-0563/

import math
from typing import Tuple
from typing_extensions import Literal

from multiformats import varint

from ..ipld import IPLDKind
from ..encoding import _dag_cbor_code
from .err import CBORDecodingError
from ._stream import Stream, StreamSnapshot
from ._err_utils import _bytes2hex, _decode_error_msg_lines, _decode_error_msg, _extract_error_cause_lines, _cid_error_template

def _required_multicodec(stream: Stream) -> str:
    curr_snapshot = stream.curr_snapshot
    msg = "Required 'dag-cbor' multicodec code."
    exp_bs = varint.encode(_dag_cbor_code)
    details = f"byte{'s' if curr_snapshot.latest_read_size > 1 else ''} should be 0x{exp_bs.hex()}."
    return _decode_error_msg(msg, curr_snapshot, details=details)

def _multiple_top_level_items(stream: Stream) -> str:
    msg = "Encode and decode must operate on a single top-level CBOR object."
    details = "unexpected start byte of a second top-level CBOR object"
    return _decode_error_msg(msg, stream.curr_snapshot, details=details)

def _invalid_float(stream: Stream, arg: float) -> str:
    if math.isnan(arg):
        msg = "NaN is not an allowed float value."
        float_str = "float('NaN')"
    else:
        assert math.isinf(arg), "Float must be NaN or infinite."
        s = ("" if arg > 0 else "-")
        msg = s+"Infinity is not an allowed float value."
        float_str = f"float('{s}Infinity')"
    details = f"struct.pack('>d', {float_str})"
    return _decode_error_msg(msg, stream.curr_snapshot, details=details, hl_start=1)

def _unexpected_eof(stream: Stream, what: str, n: int, include_prev_snapshot: bool = True) -> str:
    prev_snapshot = stream.prev_snapshot if include_prev_snapshot else StreamSnapshot(bytes(), 0)
    curr_snapshot = stream.curr_snapshot
    msg = f"Unexpected EOF while attempting to read {what}."
    bytes_read = curr_snapshot.latest_read_size
    hl_start = prev_snapshot.latest_read_size
    details = f"{bytes_read} bytes read, out of {n} expected."
    snapshots = [prev_snapshot, curr_snapshot] if include_prev_snapshot else [curr_snapshot]
    return _decode_error_msg(msg, *snapshots, details=details, hl_start=hl_start)

def _invalid_additional_info(stream: Stream, additional_info: int, major_type: int) -> str:
    msg = f"Invalid additional info {additional_info} in data item head for major type 0x{major_type:x}."
    if major_type == 0x7:
        details = f"lower 5 bits are {additional_info:0>5b}, expected from {0:0>5b} to {23:0>5b}, or {27:0>5b}."
    else:
        details = f"lower 5 bits are {additional_info:0>5b}, expected from {0:0>5b} to {27:0>5b}."
    return _decode_error_msg(msg, stream.curr_snapshot, details=details)

def _excessive_int_size(stream: Stream, arg: int, bytes_used: int, bytes_sufficient: int) -> str:
    s = 's' if bytes_sufficient > 1 else ''
    msg = f"Integer {arg} was encoded using {bytes_used} bytes, while {bytes_sufficient} byte{s} would have been enough."
    details = f"same as byte{s} 0x{arg:0>{2*bytes_sufficient}x}"
    return _decode_error_msg(msg, stream.prev_snapshot, stream.curr_snapshot, details=details, hl_start=1)

def _unicode(stream: Stream, length: int, start: int, end: int, reason: str) -> str:
    prev_snapshot = stream.prev_snapshot
    curr_snapshot = stream.curr_snapshot
    msg = "String bytes are not valid utf-8 bytes."
    lines = [msg]
    str_details = f"string of length {length}"
    lines.extend(_decode_error_msg_lines(prev_snapshot, curr_snapshot, details=str_details, hl_len=1))
    lines.extend(_decode_error_msg_lines(curr_snapshot, details=reason, start=start, end=end, pad_start=start+prev_snapshot.latest_read_size))
    return "\n".join(lines)

def _list_item(list_head_snapshot: StreamSnapshot, idx: int, length: int, e: CBORDecodingError) -> str:
    msg = "Error while decoding list."
    lines = [
        msg,
        *_decode_error_msg_lines(list_head_snapshot, details=f"list of length {length}", dots=True),
        f"Error occurred while decoding item at position {idx}: further details below.",
        *_extract_error_cause_lines(e)
    ]
    return "\n".join(lines)

def _dict_key_type(stream: Stream, major_type: int) -> str:
    msg = "Dictionary key is not of string type."
    details = f"major type is {hex(major_type)}, should be 0x3 (string) instead."
    return _decode_error_msg(msg, stream.curr_snapshot, details=details, hl_len=1, dots=True)

def _dict_item(dict_head_snapshot: StreamSnapshot, item: Literal["key", "value"], idx: int, length: int, e: CBORDecodingError) -> str:
    msg = "Error while decoding dict."
    details = f"dict of length {length}"
    lines = [
        msg,
        *_decode_error_msg_lines(dict_head_snapshot, details=details, dots=True),
        f"Error occurred while decoding {item} at position {idx}: further details below.",
        *_extract_error_cause_lines(e)
    ]
    return "\n".join(lines)

def _duplicate_dict_key(dict_head_snapshot: StreamSnapshot, stream: Stream, k: str, idx: int, length: int) -> str:
    msg = "Error while decoding dict."
    dict_details = f"dict of length {length}"
    key_details = f"decodes to key {repr(k)}"
    lines = [
        msg,
        *_decode_error_msg_lines(dict_head_snapshot, details=dict_details, dots=True),
        f"Duplicate key is found at position {idx}.",
        *_decode_error_msg_lines(stream.curr_snapshot, details=key_details)
    ]
    return "\n".join(lines)

def _dict_key_order(dict_head_snapshot: StreamSnapshot, kb0: bytes, idx0: int, kb1: bytes, idx1: int, length: int) -> str:
    # pylint: disable = too-many-arguments
    msg = "Error while decoding dict."
    pad_len = max(len(str(idx0)), len(str(idx1)))
    idx0_str = f"{idx0: >{pad_len}}"
    idx1_str = f"{idx1: >{pad_len}}"
    details = f"dict of length {length}"
    lines = [
        msg,
        *_decode_error_msg_lines(dict_head_snapshot, details=details, dots=True),
        "Dictionary keys not in canonical order.",
        f"  Key at pos #{idx0_str}: {_bytes2hex(kb0)}",
        f"  Key at pos #{idx1_str}: {_bytes2hex(kb1)}",
    ]
    return "\n".join(lines)

def _invalid_tag(stream: Stream, arg: int) -> str:
    prev_snapshot = stream.prev_snapshot
    curr_snapshot = stream.curr_snapshot
    msg = "Error while decoding item of major type 0x6: only tag 42 is allowed."
    details = f"tag {arg}"
    hl_start = prev_snapshot.latest_read_size
    return _decode_error_msg(msg, prev_snapshot, curr_snapshot, details=details, hl_start=hl_start)

def _cid(cid_head_snapshots: Tuple[StreamSnapshot, StreamSnapshot], e: CBORDecodingError) -> str:
    return _cid_error_template(cid_head_snapshots, *_extract_error_cause_lines(e))

def _cid_bytes(cid_head_snapshots: Tuple[StreamSnapshot, StreamSnapshot], stream: Stream, cid_bytes: IPLDKind) -> str:
    decoded_type = type(cid_bytes).__name__
    details = f"decodes to an item of type {repr(decoded_type)}"
    explanation = [
        "CID bytes did not decode to an item of type 'bytes'.",
        *_decode_error_msg_lines(stream.curr_snapshot, details=details),
    ]
    return _cid_error_template(cid_head_snapshots, *explanation)

def _cid_multibase(cid_head_snapshots: Tuple[StreamSnapshot, StreamSnapshot], stream: Stream, cid_bytes: bytes) -> str:
    details = "byte should be 0x00"
    explanation = [
        "CID does not start with the identity Multibase prefix.",
        *_decode_error_msg_lines(stream.prev_snapshot, stream.curr_snapshot, details=details, hl_start=1, hl_len=1),
    ]
    return _cid_error_template(cid_head_snapshots, *explanation)

def _simple_value(stream: Stream, arg: int) -> str:
    msg = "Error while decoding major type 0x7: allowed simple values are 0x14, 0x15 and 0x16."
    details = f"simple value is {arg}"
    return _decode_error_msg(msg, stream.curr_snapshot, details=details)



"""
    A script to generate .rst files for API documentation.
"""

import glob
import importlib
import inspect
import json
import os
import pkgutil
from typing import Dict, List, Optional, Tuple
import sys

from typing_validation import validate

def _list_package_contents(pkg_name: str) -> List[str]:
    modules = [pkg_name]
    for submod in pkgutil.iter_modules([pkg_name.replace(".", "/")]):
        submod_fullname = pkg_name+"."+submod.name
        if submod.ispkg:
            for subsubmod_name in _list_package_contents(submod_fullname):
                modules.append(subsubmod_name)
        else:
            modules.append(submod_fullname)
    return modules

def make_apidocs() -> None:
    """
        A script to generate .rst files for API documentation.
    """
    err_msg = """Expected a 'make-api.json' file, with the following structure:
{
    "pkg_name": str,
    "apidocs_folder": str,
    "pkg_path": str,
    "toc_filename": str,
    "type_alias_dict_filename": Optional[str],
    "include_members": Dict[str, List[str]],
    "type_aliases": Dict[str, List[str]],
    "exclude_members": Dict[str, List[str]],
    "exclude_modules": List[str],
    "member_fullnames": Dict[str, Dict[str, str]],
    "special_class_members": Dict[str, List[str]],
}

Set "toc_filename" to null to avoid generating a table of contents file.

"""
    try:
        with open("make-api.json", "r") as f:
            config = json.load(f)
            pkg_name = config.get("pkg_name", None)
            validate(pkg_name, str)
            pkg_path = config.get("pkg_path", None)
            validate(pkg_path, str)
            apidocs_folder = config.get("apidocs_folder", None)
            validate(apidocs_folder, str)
            toc_filename = config.get("toc_filename", None)
            validate(toc_filename, str)
            type_alias_dict_filename = config.get("type_alias_dict_filename", None)
            validate(type_alias_dict_filename, Optional[str])
            include_members = config.get("include_members", {})
            validate(include_members, Dict[str, List[str]])
            type_aliases = config.get("type_aliases", {})
            validate(type_aliases, Dict[str, List[str]])
            exclude_members = config.get("exclude_members", {})
            validate(exclude_members, Dict[str, List[str]])
            include_modules = config.get("include_modules", [])
            validate(include_modules, List[str])
            exclude_modules = config.get("exclude_modules", [])
            validate(exclude_modules, List[str])
            member_fullnames = config.get("member_fullnames", {})
            validate(member_fullnames, Dict[str, Dict[str, str]])
            special_class_members = config.get("special_class_members", {})
            validate(special_class_members, Dict[str, List[str]])
    except FileNotFoundError:
        print(err_msg)
        sys.exit(1)
    except TypeError:
        print(err_msg)
        sys.exit(1)
    for mod_name, type_alias_members in type_aliases.items():
        if mod_name not in include_members:
            include_members[mod_name] = []
        include_members[mod_name].extend(type_alias_members)

    cwd = os.getcwd()
    os.chdir(pkg_path)
    sys.path = [os.getcwd()]+sys.path
    modules = _list_package_contents(pkg_name)
    modules_dict = {
        mod_name: importlib.import_module(mod_name)
        for mod_name in modules
    }
    for mod_name in include_modules:
        if mod_name not in modules_dict:
            modules_dict[mod_name] = importlib.import_module(mod_name)
    os.chdir(cwd)

    print(f"Removing all docfiles from {apidocs_folder}/")
    for apidoc_file in glob.glob(f"{apidocs_folder}/*.rst"):
        print(f"    {apidoc_file}")
        os.remove(apidoc_file)
    print()

    type_alias_fullnames: dict[str, str] = {}

    print("Pre-processing type aliases:")
    for mod_name, mod_type_aliases in type_aliases.items():
        if mod_name in exclude_modules:
            continue
        for member_name in mod_type_aliases:
            member_fullname = f"{mod_name}.{member_name}"
            if member_name in type_alias_fullnames:
                print(f"    WARNING! Skipping type alias {member_name} -> {member_fullname}")
                print(f"             Existing type alias {member_name} -> {type_alias_fullnames[member_name]}")
            else:
                type_alias_fullnames[member_name] = member_fullname
                print(f"    {member_name} -> {member_fullname}")
    print()

    for mod_name, mod in modules_dict.items():
        if mod_name in exclude_modules:
            continue
        filename = f"{apidocs_folder}/{mod_name}.rst"
        print(f"Writing API docfile {filename}")
        lines: List[str] = [
            mod_name,
            "="*len(mod_name),
            "",
            f".. automodule:: {mod_name}",
            ""
        ]
        mod__all__ = getattr(mod, "__all__", [])
        reexported_members: List[Tuple[str, str]] = []
        for member_name in sorted(name for name in dir(mod)):
            to_include = mod_name in include_members and member_name in include_members[mod_name]
            to_exclude = mod_name in exclude_members and member_name in exclude_members[mod_name]
            if to_exclude:
                continue
            if member_name.startswith("_") and not to_include:
                continue
            member = getattr(mod, member_name)
            member_module = inspect.getmodule(member)
            member_module_name = member_module.__name__ if member_module is not None else None
            imported_member = member_module is not None and member_module != mod
            if mod_name in include_members and member_name in include_members[mod_name]:
                imported_member = False
            if member_name in type_alias_fullnames:
                member_fullname = type_alias_fullnames[member_name]
            elif mod_name in member_fullnames and member_name in member_fullnames[mod_name]:
                member_fullname = member_fullnames[mod_name][member_name]
            elif imported_member:
                if inspect.ismodule(member):
                    member_fullname = member_module_name or ""
                else:
                    member_fullname = f"{member_module_name}.{member_name}"
            else:
                member_fullname = f"{mod_name}.{member_name}"
            member_kind = "data"
            if inspect.isclass(member):
                member_kind = "class"
            elif inspect.isfunction(member):
                member_kind = "function"
            elif inspect.ismodule(member):
                member_kind = "module"
            if not imported_member:
                member_lines: List[str] = []
                member_lines = [
                    member_name,
                    "-"*len(member_name),
                    "",
                    f".. auto{member_kind}:: {member_fullname}",
                ]
                if member_kind == "class":
                    member_lines.append("    :show-inheritance:")
                    member_lines.append("    :members:")
                    if member_fullname in special_class_members and special_class_members[member_fullname]:
                        member_lines.append(f"    :special-members: {', '.join(special_class_members[member_fullname])}")
                member_lines.append("")
                if member_name in type_alias_fullnames:
                    print(f"    {member_kind} {member_name} -> {type_alias_fullnames[member_name]} (type alias)")
                else:
                    print(f"    {member_kind} {member_name}")
                lines.extend(member_lines)
            elif member_name in mod__all__:
                reexported_members.append((member_fullname, member_kind))
        if reexported_members:
            reexported_members_header = f"{mod_name}.__all__"
            print(f"    {reexported_members_header}:")
            lines.extend([
                reexported_members_header,
                "-"*len(reexported_members_header),
                "",
                "The following members were explicitly reexported using ``__all__``:",
                "",
            ])
            refkinds = {
                "data": "obj",
                "function": "func",
                "class": "class",
                "module": "mod"
            }
            for member_fullname, member_kind in reexported_members:
                refkind = f":py:{refkinds[member_kind]}:"
                lines.append(f"    - {refkind}`{member_fullname}`")
                print(f"        {member_kind} {member_fullname}")
            lines.append("")
        with open(filename, "w") as f:
            f.write("\n".join(lines))
        print("")

    toctable_lines = [
        ".. toctree::",
        "    :maxdepth: 2",
        "    :caption: API Documentation",
        ""
    ]
    print(f"Writing TOC for API docfiles at {toc_filename}")
    for mod_name in modules_dict:
        if mod_name in exclude_modules:
            continue
        line = f"    {apidocs_folder}/{mod_name}"
        toctable_lines.append(line)
        print(line)
    toctable_lines.append("")
    print()

    with open(toc_filename, "w") as f:
        f.write("\n".join(toctable_lines))

    if type_alias_dict_filename is not None:
        print(f"Writing type alias dictionary: {type_alias_dict_filename}")
        for name, fullname in type_alias_fullnames.items():
            print(f"    {name} -> {fullname}")
        print()
        with open(type_alias_dict_filename, "w") as f:
            json.dump(type_alias_fullnames, f, indent=4)

if __name__ == "__main__":
    make_apidocs()



r"""
    Errors for the :mod:`dag_cbor.decoding` submodule.
"""

from ..encoding.err import CBORError, DAGCBORError

class CBORDecodingError(CBORError):
    """
        Class for decoding errors due to the CBOR specification.
    """

class DAGCBORDecodingError(CBORDecodingError, DAGCBORError):
    """
        Class for decoding errors due to the DAG-CBOR specification.
    """



"""
    Functions to generate random data.
"""
# pylint: disable = global-statement

from __future__ import annotations # See https://peps.python.org/pep-0563/

from contextlib import contextmanager
import math
from random import Random # pylint: disable = import-self
import sys
from types import MappingProxyType
from typing import Any, Dict, Iterator, List, Mapping, Optional, Tuple
from typing_validation import validate

from multiformats import multicodec, multibase, multihash, CID

from .ipld import IPLDKind
from .encoding import encode, _canonical_order_dict

_min_int = -18446744073709551616
_max_int = 18446744073709551615
_min_float = -sys.float_info.max
_max_float = sys.float_info.max
_min_codepoint = 0x00
_max_codepoint = 0x10FFFF

_default_options: Dict[str, Any] = {
    "min_int": -100,
    "max_int": 100,
    "min_bytes": 0,
    "max_bytes": 8,
    "min_chars": 0,
    "max_chars": 8,
    "min_codepoint": 0x21,
    "max_codepoint": 0x7e,
    "min_len": 0,
    "max_len": 8,
    "max_nesting": 2,
    "canonical": True,
    "min_float": -100.0,
    "max_float": 100.0,
    "float_decimals": 3,
    "include_cid": True,
}

_options = _default_options
_rand = Random(0)

def reset_options() -> None:
    """
        Resets random generation options to their default values.
    """
    global _options
    global _rand
    _options = _default_options
    _rand = Random(0)

def default_options() -> Mapping[str, Any]:
    """
        Readonly view of the default random generation options.
    """
    return MappingProxyType(_default_options)

def get_options() -> Mapping[str, Any]:
    """
        Readonly view of the current random generation options.
    """
    return MappingProxyType(_options)

@contextmanager
def options(*,
            seed: Optional[int] = None,
            min_int: Optional[int] = None,
            max_int: Optional[int] = None,
            min_bytes: Optional[int] = None,
            max_bytes: Optional[int] = None,
            min_chars: Optional[int] = None,
            max_chars: Optional[int] = None,
            min_codepoint: Optional[int] = None,
            max_codepoint: Optional[int] = None,
            min_len: Optional[int] = None,
            max_len: Optional[int] = None,
            max_nesting: Optional[int] = None,
            canonical: Optional[bool] = None,
            min_float: Optional[float] = None,
            max_float: Optional[float] = None,
            float_decimals: Optional[int] = None,
            include_cid: Optional[bool] = None,) -> Iterator[None]:
    """
        Returns with-statement context manager for temporary option setting:

        .. code-block:: python

            with options(**options):
                for value in rand_data(num_samples):
                    ...

        Options available:

        .. code-block::

            seed: int           # set new random number generator, with this seed
            min_int: int        # smallest `int` value
            max_int: int        # largest `int` value
            min_bytes: int      # min length of `bytes` value
            max_bytes: int      # max length of `bytes` value
            min_chars: int      # min length of `str` value
            max_chars: int      # max length of `str` value
            min_codepoint: int  # min utf-8 codepoint in `str` value
            max_codepoint: int  # max utf-8 codepoint in `str` value
            min_len: int        # min length of `list` and `dict` values
            max_len: int        # max length of `list` and `dict` values
            max_nesting: int    # max nesting of collections
            canonical: bool     # whether `dict` values have canonically ordered keys
            min_float: float    # smallest `float` value
            max_float: float    # largest `float` value
            float_decimals: int # number of decimals to keep in floats
            include_cid: bool   # whether to generate CID values
    """
    # pylint: disable = too-many-locals, too-many-arguments
    global _options
    global _rand
    _old_options = _options
    _old_rand = _rand
    try:
        set_options(seed=seed,
                    min_int=min_int, max_int=max_int,
                    min_bytes=min_bytes, max_bytes=max_bytes,
                    min_chars=min_chars, max_chars=max_chars,
                    min_codepoint=min_codepoint, max_codepoint=max_codepoint,
                    min_len=min_len, max_len=max_len,
                    max_nesting=max_nesting, canonical=canonical,
                    min_float=min_float, max_float=max_float,
                    float_decimals=float_decimals, include_cid=include_cid)
        yield
    finally:
        _options = _old_options
        _rand = _old_rand

def set_options(*,
                seed: Optional[int] = None,
                min_int: Optional[int] = None,
                max_int: Optional[int] = None,
                min_bytes: Optional[int] = None,
                max_bytes: Optional[int] = None,
                min_chars: Optional[int] = None,
                max_chars: Optional[int] = None,
                min_codepoint: Optional[int] = None,
                max_codepoint: Optional[int] = None,
                min_len: Optional[int] = None,
                max_len: Optional[int] = None,
                max_nesting: Optional[int] = None,
                canonical: Optional[bool] = None,
                min_float: Optional[float] = None,
                max_float: Optional[float] = None,
                float_decimals: Optional[int] = None,
                include_cid: Optional[bool] = None,) -> None:
    """
        Permanently sets random generation options. See :func:`options` for the available options.

    """
    # pylint: disable = too-many-branches, too-many-locals, too-many-statements, too-many-arguments
    for iarg in (seed, min_int, max_int, min_bytes, max_bytes, min_chars, max_chars,
                 min_codepoint, max_codepoint, min_len, max_len, max_nesting, float_decimals):
        validate(iarg, Optional[int])
    for barg in (canonical, include_cid):
        validate(barg, Optional[bool])
    for farg in (min_float, max_float):
        validate(farg, Optional[float])
    global _options
    global _rand
    # set newly passed options
    _new_options: Dict[str, Any] = {}
    if seed is not None:
        _rand = Random(seed)
    if min_int is not None:
        if min_int < _min_int:
            raise ValueError("Value for min_int is not a valid CBOR integer.")
        _new_options["min_int"] = min_int
    if max_int is not None:
        if max_int > _max_int:
            raise ValueError("Value for max_int is not a valid CBOR integer.")
        _new_options["max_int"] = max_int
    if min_bytes is not None:
        if min_bytes < 0:
            raise ValueError("Value for min_bytes is negative.")
        _new_options["min_bytes"] = min_bytes
    if max_bytes is not None:
        if max_bytes < 0:
            raise ValueError("Value for max_bytes is negative.")
        _new_options["max_bytes"] = max_bytes
    if min_chars is not None:
        if min_chars < 0:
            raise ValueError("Value for min_chars is negative.")
        _new_options["min_chars"] = min_chars
    if max_chars is not None:
        if max_chars < 0:
            raise ValueError("Value for max_chars is negative.")
        _new_options["max_chars"] = max_chars
    if min_codepoint is not None:
        if min_codepoint < _min_codepoint or min_codepoint > _max_codepoint:
            raise ValueError("Value for min_codepoint not a valid utf-8 codepoint.")
        _new_options["min_codepoint"] = min_codepoint
    if max_codepoint is not None:
        if max_codepoint < _min_codepoint or max_codepoint > _max_codepoint:
            raise ValueError("Value for max_codepoint not a valid utf-8 codepoint.")
        _new_options["max_codepoint"] = max_codepoint
    if min_len is not None:
        if min_len < 0:
            raise ValueError("Value for min_len is negative.")
        _new_options["min_len"] = min_len
    if max_len is not None:
        if max_len < 0:
            raise ValueError("Value for max_len is negative.")
        _new_options["max_len"] = max_len
    if max_nesting is not None:
        if max_nesting < 0:
            raise ValueError("Value for max_nesting is negative.")
        _new_options["max_nesting"] = max_nesting
    if canonical is not None:
        _new_options["canonical"] = canonical
    if min_float is not None:
        if math.isnan(min_float) or math.isinf(min_float):
            raise ValueError("Value for min_float is not a valid CBOR float.")
        _new_options["min_float"] = min_float
    if max_float is not None:
        if math.isnan(max_float) or math.isinf(max_float):
            raise ValueError("Value for max_float is not a valid CBOR float.")
        _new_options["max_float"] = max_float
    if float_decimals is not None:
        if float_decimals < 0:
            raise ValueError("Value for float_decimals is negative.")
        _new_options["float_decimals"] = float_decimals
    if include_cid is not None:
        _new_options["include_cid"] = include_cid
    # pass-through other options with former values
    for k, v in _options.items():
        if k not in _new_options:
            _new_options[k] = v
    # check compatibility conditions
    if _new_options["min_bytes"] > _new_options["max_bytes"]:
        raise ValueError("Value for min_bytes is larger than value for max_bytes.")
    if _new_options["min_chars"] > _new_options["max_chars"]:
        raise ValueError("Value for min_chars is larger than value for max_chars.")
    if _new_options["min_codepoint"] > _new_options["max_codepoint"]:
        raise ValueError("Value for min_codepoint is larger than value for max_codepoint.")
    if _new_options["min_len"] > _new_options["max_len"]:
        raise ValueError("Value for min_len is larger than value for max_len.")
    # update options
    _options = _new_options


def rand_data(n: Optional[int] = None, *, max_nesting: Optional[int] = None) -> Iterator[IPLDKind]:
    r"""
        Generates a stream of random data.

        :param n: the number of samples to be yielded; if :obj:`None`, an infinite stream is yielded
        :param max_nesting: the maximum nesting level for containers; if :obj:`None`, value from :func:`get_options` is used

        Maximum nesting level for containers:

        - the integer value -1 means no containers will be generated
        - integer values >= 0 mean that containers will be generated, with items generated by ``random_data(max_nesting=max_nesting-1)``
        - no other values are valid

    """
    validate(n, Optional[int])
    validate(max_nesting, Optional[int])
    return _rand_data(n, max_nesting=max_nesting)

def _rand_data(n: Optional[int] = None, *, max_nesting: Optional[int] = None) -> Iterator[IPLDKind]:
    if n is not None and n < 0:
        raise ValueError()
    if max_nesting is None:
        max_nesting = _options["max_nesting"]
    elif max_nesting < -1:
        raise ValueError("Value for max_nesting must be >= -1 (with -1 indicating no containers).")
    include_cid = _options["include_cid"]
    data_generators: List[Iterator[Any]] = [
        _rand_list(max_nesting=max_nesting) if max_nesting >= 0 else iter([]),
        _rand_dict(max_nesting=max_nesting) if max_nesting >= 0 else iter([]),
        _rand_int(),
        _rand_bytes(),
        _rand_str(),
        _rand_bool_none(),
        _rand_float(),
        _rand_cid()
    ]
    num_data_generators = len(data_generators) if include_cid else len(data_generators)-1
    i = 0
    while n is None or i < n:
        if max_nesting == -1:
            # exclude containers
            datatype = _rand.randrange(0x2, num_data_generators)
        else:
            # include containers
            datatype = _rand.randrange(0x0, num_data_generators)
        try:
            yield next(data_generators[datatype])
        except StopIteration as e:
            raise RuntimeError("All random streams are infinite, this should not happen.") from e
        i += 1

def rand_list(n: Optional[int] = None, *, length: Optional[int] = None, max_nesting: Optional[int] = None) -> Iterator[List[Any]]:
    """
        Generates a stream of random :obj:`list` data.

        :param n: the number of samples to be yielded; if :obj:`None`, an infinite stream is yielded
        :param length: size for the lists; if :obj:`None`, a random value is sampled according to the :func:`options`
        :param max_nesting: the maximum nesting level for containers; if :obj:`None`, value from :func:`get_options` is used
    """
    validate(n, Optional[int])
    validate(length, Optional[int])
    validate(max_nesting, Optional[int])
    return _rand_list(n, length=length, max_nesting=max_nesting)

def _rand_list(n: Optional[int] = None, *, length: Optional[int] = None, max_nesting: Optional[int] = None) -> Iterator[List[Any]]:
    if n is not None and n < 0:
        raise ValueError()
    if length is not None and length < 0:
        raise ValueError()
    if max_nesting is None:
        max_nesting = _options["max_nesting"]
    elif max_nesting < 0:
        raise ValueError("Value for max_nesting is negative.")
    min_len = _options["min_len"]
    max_len = _options["max_len"]
    i = 0
    while n is None or i < n:
        _length = length if length is not None else _rand.randint(min_len, max_len)
        yield list(_rand_data(_length, max_nesting=max_nesting-1))
        i += 1

def rand_dict(n: Optional[int] = None, *, length: Optional[int] = None, max_nesting: Optional[int] = None) -> Iterator[Dict[str, Any]]:
    """
        Generates a stream of random :obj:`dict` data.

        :param n: the number of samples to be yielded; if :obj:`None`, an infinite stream is yielded
        :param length: size for the dicts; if :obj:`None`, a random value is sampled according to the :func:`options`
        :param max_nesting: the maximum nesting level for containers; if :obj:`None`, value from :func:`get_options` is used
    """
    validate(n, Optional[int])
    validate(length, Optional[int])
    validate(max_nesting, Optional[int])
    return _rand_dict(n, length=length, max_nesting=max_nesting)

def _rand_dict(n: Optional[int] = None, *, length: Optional[int] = None, max_nesting: Optional[int] = None) -> Iterator[Dict[str, Any]]:
    # pylint: disable = too-many-locals, too-many-branches
    if n is not None and n < 0:
        raise ValueError()
    if length is not None and length < 0:
        raise ValueError()
    if max_nesting is None:
        max_nesting = _options["max_nesting"]
    elif max_nesting < 0:
        raise ValueError("Value for max_nesting is negative.")
    min_len = _options["min_len"]
    max_len = _options["max_len"]
    canonical = _options["canonical"]
    min_chars = _options["min_chars"]
    max_chars = _options["max_chars"]
    max_codepoint = _options["max_codepoint"]
    num_codepoints = max_codepoint-_options["min_codepoint"]
    i = 0
    while n is None or i < n:
        _length = length if length is not None else _rand.randint(min_len, max_len)
        # check whether we have enough distinct strings to generate a random dictionary of desired length
        if num_codepoints == 1:
            num_strings = max_chars-min_chars+1
        else:
            num_strings = (num_codepoints**min_chars)*(num_codepoints**(max_chars-min_chars+1)-1)//(num_codepoints-1)
        if num_strings < _length:
            raise ValueError(f"Not enough distinct strings available to make a dictionary of length {_length}")
        # generate distinct dictionary keys
        if num_codepoints == 1:
            key_lengths = _rand.sample(range(min_chars, max_chars+1), _length)
            keys = [chr(max_codepoint)*l for l in key_lengths]
        else:
            keys = []
            keys_set = set()
            str_generator = _rand_str()
            while len(keys) < _length:
                try:
                    s = next(str_generator)
                except StopIteration as e:
                    raise RuntimeError("Random string stream is infinite, this should not happen.") from e
                if s not in keys_set:
                    keys.append(s)
                    keys_set.add(s)
        # generate dictionary
        raw_dict = dict(zip(keys, _rand_data(_length, max_nesting=max_nesting-1)))
        if canonical:
            yield _canonical_order_dict(raw_dict)
        else:
            yield raw_dict
        i += 1

def rand_int(n: Optional[int] = None) -> Iterator[int]:
    """
        Generates a stream of random :obj:`int` data.

        :param n: the number of samples to be yielded; if :obj:`None`, an infinite stream is yielded
    """
    validate(n, Optional[int])
    return _rand_int(n)

def _rand_int(n: Optional[int] = None) -> Iterator[int]:
    if n is not None and n < 0:
        raise ValueError()
    min_int = _options["min_int"]
    max_int = _options["max_int"]
    i = 0
    while n is None or i < n:
        yield _rand.randint(min_int, max_int)
        i += 1

def rand_bytes(n: Optional[int] = None, *, length: Optional[int] = None) -> Iterator[bytes]:
    """
        Generates a stream of random :obj:`bytes` data.

        :param n: the number of samples to be yielded; if :obj:`None`, an infinite stream is yielded
        :param length: length of the bytestrings; if :obj:`None`, a random value is sampled according to the :func:`options`
    """
    validate(n, Optional[int])
    validate(length, Optional[int])
    return _rand_bytes(n, length=length)

def _rand_bytes(n: Optional[int] = None, *, length: Optional[int] = None) -> Iterator[bytes]:
    if n is not None and n < 0:
        raise ValueError()
    if length is not None and length < 0:
        raise ValueError()
    min_bytes = _options["min_bytes"]
    max_bytes = _options["max_bytes"]
    i = 0
    while n is None or i < n:
        _length = length if length is not None else _rand.randint(min_bytes, max_bytes)
        yield bytes([_rand.randint(0, 255) for _ in range(_length)])
        i += 1

def rand_str(n: Optional[int] = None, *, length: Optional[int] = None) -> Iterator[str]:
    """
        Generates a stream of random :obj:`str` data.

        :param n: the number of samples to be yielded; if :obj:`None`, an infinite stream is yielded
        :param length: length of the strings; if :obj:`None`, a random value is sampled according to the :func:`options`
    """
    validate(n, Optional[int])
    validate(length, Optional[int])
    return _rand_str(n, length=length)

def _rand_str(n: Optional[int] = None, *, length: Optional[int] = None) -> Iterator[str]:
    if n is not None and n < 0:
        raise ValueError()
    if length is not None and length < 0:
        raise ValueError()
    min_chars = _options["min_chars"]
    max_chars = _options["max_chars"]
    min_codepoint = _options["min_codepoint"]
    max_codepoint = _options["max_codepoint"]
    i = 0
    while n is None or i < n:
        _length = length if length is not None else _rand.randint(min_chars, max_chars)
        codepoints = [_rand.randint(min_codepoint, max_codepoint) for _ in range(_length)]
        try:
            string = "".join(chr(c) for c in codepoints)
            string.encode("utf-8", errors="strict")
            yield string
            i += 1
        except UnicodeError:
            continue

def rand_bool(n: Optional[int] = None) -> Iterator[bool]:
    """
        Generates a stream of random :obj:`bool` data.

        :param n: the number of samples to be yielded; if :obj:`None`, an infinite stream is yielded
    """
    validate(n, Optional[int])
    return _rand_bool(n)

def _rand_bool(n: Optional[int] = None) -> Iterator[bool]:
    if n is not None and n < 0:
        raise ValueError()
    i = 0
    while n is None or i < n:
        x = _rand.randint(0, 1)
        yield x == 1
        i += 1

def rand_bool_none(n: Optional[int] = None) -> Iterator[Optional[bool]]:
    """
        Generates a stream of random :obj:`bool` or :obj:`None` data.

        :param n: the number of samples to be yielded; if :obj:`None`, an infinite stream is yielded
    """
    validate(n, Optional[int])
    return _rand_bool_none(n)

def _rand_bool_none(n: Optional[int] = None) -> Iterator[Optional[bool]]:
    if n is not None and n < 0:
        raise ValueError()
    i = 0
    while n is None or i < n:
        x = _rand.randint(0, 2)
        yield None if x == 2 else x == 1
        i += 1

def rand_float(n: Optional[int] = None) -> Iterator[float]:
    """
        Generates a stream of random :obj:`float` data.

        :param n: the number of samples to be yielded; if :obj:`None`, an infinite stream is yielded
    """
    validate(n, Optional[int])
    return _rand_float(n)

def _rand_float(n: Optional[int] = None) -> Iterator[float]:
    if n is not None and n < 0:
        raise ValueError()
    min_float = _options["min_float"]
    max_float = _options["max_float"]
    float_decimals = _options["float_decimals"]
    eps = 10.0**-float_decimals
    if min_float >= 0 or max_float <= 0:
        # no overflow in `min_float + (max_float-min_float) * random()`, can use `Random.uniform`
        i = 0
        while n is None or i < n:
            x = _rand.uniform(min_float, max_float)
            yield x-x%eps
            i += 1
    else:
        # overflow in `min_float + (max_float-min_float) * random()`, cannot use `Random.uniform`
        i = 0
        while n is None or i < n:
            x = 1/(1+max_float/(-min_float))
            # x is (-min_float)/(max_float-min_float), the probability of sampling a number in (-min_float, 0)
            if _rand.random() < x:
                x = _rand.random()*min_float
            else:
                x = _rand.random()*max_float
            yield x-x%eps
            i += 1

_cid_multibase = multibase.get("base58btc") # the default base for binary CIDs
_cid_version = 1
_cid_multicodec = multicodec.get("dag-cbor")
_cid_multihash = multihash.get("sha3-512")

def rand_data_cid(n: Optional[int] = None, *, max_nesting: Optional[int] = None) -> Iterator[Tuple[IPLDKind, CID]]:
    r"""
        Generates a stream of random DAG-CBOR data and associated CIDs:

        - multibase 'base32'
        - CIDv1
        - multicodec 'dag-cbor'
        - multihash 'sha3-512', with full 512-bit digest

        :param n: the number of samples to be yielded; if :obj:`None`, an infinite stream is yielded
        :param max_nesting: the maximum nesting level for containers; if :obj:`None`, value from :func:`get_options` is used

    """
    validate(n, Optional[int])
    return _rand_data_cid(n, max_nesting=max_nesting)

def rand_cid(n: Optional[int] = None, *, max_nesting: Optional[int] = None) -> Iterator[CID]:
    """
        Generates a stream of random CIDs:

        - multibase 'base32'
        - CIDv1
        - multicodec 'dag-cbor'
        - multihash 'sha3-512', with full 512-bit digest

        :param n: the number of samples to be yielded; if :obj:`None`, an infinite stream is yielded
        :param max_nesting: the maximum nesting level for containers; if :obj:`None`, value from :func:`get_options` is used
    """
    validate(n, Optional[int])
    return _rand_cid(n, max_nesting=max_nesting)

def _rand_cid(n: Optional[int] = None, *, max_nesting: Optional[int] = None) -> Iterator[CID]:
    return (cid for _, cid in _rand_data_cid(n, max_nesting=max_nesting))

def _rand_data_cid(n: Optional[int] = None, *, max_nesting: Optional[int] = None) -> Iterator[Tuple[IPLDKind, CID]]:
    if n is not None and n < 0:
        raise ValueError()
    if max_nesting is None:
        max_nesting = _options["max_nesting"]
    elif max_nesting < 0:
        raise ValueError("Value for max_nesting is negative.")
    i = 0
    rand_data_generator = _rand_data(max_nesting=max_nesting-1)
    while n is None or i < n:
        try:
            dag_cbor_data = next(rand_data_generator)
            binary_data = encode(dag_cbor_data)
        except StopIteration as e:
            raise RuntimeError("Random digest stream is infinite, this should not happen.") from e
        yield (dag_cbor_data, CID(_cid_multibase, _cid_version, _cid_multicodec, _cid_multihash.digest(binary_data)))
        i += 1



"""
    Tests on encoding data using `dag_cbor` vs encoding data using `cbor2`.
"""
# pylint: disable = global-statement

import cbor2 # type: ignore

from dag_cbor import encode
from dag_cbor.random import rand_list, rand_dict, rand_int, rand_bytes, rand_str, rand_bool_none, rand_float, rand_cid, options

nsamples = 1000

def test_int() -> None:
    """
        Encodes random `int` samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    test_data = rand_int(nsamples)
    for i, x in enumerate(test_data):
        error_msg = f"failed at #{i} = {repr(x)}"
        assert cbor2.dumps(x) == encode(x), error_msg

def test_special_int() -> None:
    """
        Encodes specially crafted `int` samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    exponents = [8, 16, 32, 26]
    special_data = [2**e for e in exponents]
    special_data += [x-1 for x in special_data]
    special_data += [-x for x in special_data]
    for i, x in enumerate(special_data):
        error_msg = f"failed at #{i} = {repr(x)}"
        assert cbor2.dumps(x) == encode(x), error_msg

def test_bytes() -> None:
    """
        Encodes random `bytes` samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    test_data = rand_bytes(nsamples)
    for i, x in enumerate(test_data):
        error_msg = f"failed at #{i} = {repr(x)}"
        assert cbor2.dumps(x) == encode(x), error_msg

def test_str() -> None:
    """
        Encodes random `str` samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    test_data = rand_str(nsamples)
    for i, x in enumerate(test_data):
        error_msg = f"failed at #{i} = {repr(x)}"
        assert cbor2.dumps(x) == encode(x), error_msg

def test_bool_none() -> None:
    """
        Encodes random `Optional[bool]` or samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    test_data = rand_bool_none(nsamples)
    for i, x in enumerate(test_data):
        error_msg = f"failed at #{i} = {repr(x)}"
        assert cbor2.dumps(x) == encode(x), error_msg

def test_float() -> None:
    """
        Encodes random `float` samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    test_data = rand_float(nsamples)
    for i, x in enumerate(test_data):
        error_msg = f"failed at #{i} = {repr(x)}"
        assert cbor2.dumps(x) == encode(x), error_msg

def test_list() -> None:
    """
        Encodes random `list` samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    with options(include_cid=False):
        test_data = rand_list(nsamples)
        for i, x in enumerate(test_data):
            error_msg = f"failed at #{i} = {repr(x)}"
            assert cbor2.dumps(x) == encode(x), error_msg

def test_dict() -> None:
    """
        Encodes random `dict` samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    with options(include_cid=False):
        test_data = rand_dict(nsamples)
        for i, x in enumerate(test_data):
            error_msg = f"failed at #{i} = {repr(x)}"
            assert cbor2.dumps(x) == encode(x), error_msg


def test_dict_noncanonical() -> None:
    """
        Encodes a dict given in noncanonical order and tests if it is encoded in canonical order.
        from the specs (https://ipld.io/specs/codecs/dag-cbor/spec/#strictness):

        If two keys have different lengths, the shorter one sorts earlier;
        If two keys have the same length, the one with the lower value in (byte-wise) lexical order sorts earlier.
    """
    test_data = {"bar123": 5, "zap": 7, "abc432": 9}
    assert list(cbor2.loads(encode(test_data))) == ["zap", "abc432", "bar123"]


def test_cid() -> None:
    """
        Encodes random CID samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    test_data = rand_cid(nsamples)
    for i, x in enumerate(test_data):
        error_msg = f"failed at #{i} = {repr(x)}"
        assert cbor2.dumps(cbor2.CBORTag(42, b"\0" + bytes(x))) == encode(x), error_msg



r"""
    Prints error messages for a variety of decoding failures, to check that the new detailed error messages look all right.
"""
# pylint: disable = all

from typing import List
from multiformats import varint
from dag_cbor.random import rand_data
from dag_cbor import encode, decode
from dag_cbor.ipld import Kind
from dag_cbor.encoding import CBOREncodingError
from dag_cbor.decoding import CBORDecodingError

import random

random.seed(0)

test_cases = [
    # err._required_multicodec
    "00",
    "81e20301",
    # err._multiple_top_level_items
    "718301020301",
    # err._invalid_float
    "71fb7ff8000000000000",
    "71fb7ff0000000000000",
    "71fbfff0000000000000",
    # err._unexpected_eof
    "71",
    "71830102",
    "71fb3fb99999",
    "7146"+"7891bc",
    "7166"+b"hello".hex(),
    "71a1"+("66"+b"hello".hex()),
    # err._invalid_additional_info
    "715c",
    "71f9",
    # err._excessive_int_size
    "7119"+f"{156:0>4x}",
    "711a"+f"{156:0>8x}",
    "711a"+f"{32033:0>8x}",
    "711b"+f"{156:0>16x}",
    "711b"+f"{32033:0>16x}",
    "711b"+f"{2305067290:0>16x}",
    # err._unicode
    "7161"+b"\xe9".hex(),
    "7162"+b"\xe9\x80".hex(),
    "7162"+b"A\xe9".hex(),
    "7163"+b"AB\xe9".hex(),
    "7162"+b"\xe9Z".hex(),
    "7163"+b"\xe9YZ".hex(),
    "7164"+b"A\xe9YZ".hex(),
    "7165"+b"AB\xe9YZ".hex(),
    "7165"+b"AB\xe9\x80YZ".hex(),
    "7169"+b"ABCD\xe9\x80WXYZ".hex(),
    "7171"+b"ABCDEFGHIJKLMNO\xe9\x80".hex(),
    "71a1"+("63"+b"A\xe9Z".hex())+"01",
    # err._list_item
    "718401"+("1a"+f"{32033:0>8x}")+"0304",
    "718401"+("65"+b"A\xe9YZ".hex())+"0304",
    # err._dict_key_type
    "71a10101",
    "71a18301020301",
    # err._dict_item for a value
    "71a2"+("65"+b"hello".hex())+"01"+("63"+b"bye".hex())+"fb7ff0000000000000",
    # err._duplicate_dict_key for a value
    "71a3"+("65"+b"hello".hex())+"01"+("63"+b"bye".hex())+"02"+("65"+b"hello".hex())+"03",
    # err._dict_key_order
    "71a3"+("65"+b"hello".hex())+"01"+("66"+b"whatup".hex())+"02"+("63"+b"bye".hex())+"03",
    # err._invalid_tag
    "71d829"+"46"+"7891bc",
    # err._cid
    "71d82a"+"46"+"7891bc",
    # err._cid_bytes
    "71d82a"+"65"+b"hello".hex(),
    # err._cid_multibase
    "71d82a"+"450101030405",
    # err._simple_value
    "71f3"
]

def create_embedding_obj(tag: str) -> Kind:
    for obj in rand_data(max_nesting=4):
        if not isinstance(obj, dict):
            continue
        if len(obj) < 4:
            continue
        list_values = [v for v in obj.values() if isinstance(v, list) and len(v) > 4]
        if not list_values:
            continue
        l = random.choice(list_values)
        l[random.randrange(0, len(l))] = tag
        return obj
    return tag

def deep_embed(test_case: str) -> str:
    tag = "0xdeadbeef"
    obj = create_embedding_obj(tag)
    obj_bytes = encode(obj).hex()
    tag_bytes = encode(tag).hex()
    return "71"+obj_bytes.replace(tag_bytes, test_case[2:])

deep_test_cases = [
    deep_embed(random.choice(test_cases))
    for _ in range(10)
]
def print_decode_error(test_case: str) -> bool:
    encoded_bytes = bytes.fromhex(test_case)
    encoded_bytes_str = encoded_bytes.hex() if encoded_bytes else "<NO BYTES>"
    print(f"> Error raised by decoding test case {idx: >2}:\n{encoded_bytes_str}")
    print()
    try:
        decode(encoded_bytes, require_multicodec=True)
    except CBORDecodingError as e:
        print(e)
        cause = e.__cause__
        while cause is not None:
            print(cause)
            cause = cause.__cause__
        print()
        return True
    return False

if __name__ == "__main__":
    print("==== Shallow test cases ====")
    print()
    for idx, test_case in enumerate(test_cases):
        assert print_decode_error(test_case), f"Decoding of test case {idx} should have raised error."
    print("==== Deep test cases ====")
    print()
    for idx, test_case in enumerate(deep_test_cases):
        assert print_decode_error(test_case), f"Decoding of deep test case {idx} should have raised error."



"""
    Tests on number of bytes written to streams in encoding and read from streams in decoding,
    as well as tests on decoding concatenated data.
"""
# pylint: disable = global-statement

from io import BytesIO
from dag_cbor import encode, decode
from dag_cbor.ipld import IPLDKind
from dag_cbor.random import rand_data

nsamples = 1000
nconcat = 3

class BytesReadCounter:
    """ Counter for bytes read while decoding. """
    _num_bytes_read: int = 0
    def __call__(self, value: IPLDKind, num_bytes_read: int) -> None:
        self._num_bytes_read += num_bytes_read
    def __int__(self) -> int:
        return self._num_bytes_read

def test_decode_concat_length() -> None:
    """
        Encodes random item samples with `dag_cbor.encoding.encode`, then concatenates the bytes of three items.
        Decodes with `dag_cbor.decoding.decode` allowing concatenation and checks that the correct number of bytes
        are read for each encoded item.
    """
    test_data = rand_data(nconcat*nsamples)
    for i in range(nsamples):
        items = [next(test_data) for j in range(nconcat)]
        encoded_items = [encode(x) for x in items]
        stream = BytesIO(b''.join(encoded_items))
        for j in range(nconcat):
            x = items[j]
            bytes_read_cnt = BytesReadCounter()
            decode(stream, allow_concat=True, callback=bytes_read_cnt)
            error_msg = f"failed at #{i}:{j} = {repr(x)}"
            assert len(encoded_items[j]) == int(bytes_read_cnt), error_msg
        error_msg = f"failed at #{i}"
        assert len(stream.read()) == 0, error_msg

def test_encode_length() -> None:
    """
        Encodes random item samples with `dag_cbor.encoding.encode` to a stream,
        then checks that the number of bytes written is the one returned by `encode`.
    """
    test_data = rand_data(nsamples)
    for i, x in enumerate(test_data):
        error_msg = f"failed at #{i} = {repr(x)}"
        stream = BytesIO()
        num_bytes_written: int = encode(x, stream=stream)
        assert len(stream.getvalue()) == num_bytes_written, error_msg



r"""
    Types and functions relating to the IPLD data model `IPLD data model <https://ipld.io/docs/data-model/>`_.
"""

# Part of the dag-cbor library.
# Copyright (C) 2023 Hashberg Ltd

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA

from __future__ import annotations # See https://peps.python.org/pep-0563/

from typing import ClassVar, Dict, Iterator, List, MutableMapping, overload, Sequence, Tuple, Union
from weakref import WeakValueDictionary

from typing_validation import validate

from multiformats import CID

IPLDScalarKind = Union[None, bool, int, float, str, bytes, CID]
r"""
    Python type alias for scalar `kinds <https://ipld.io/docs/data-model/kinds/>`_ in the IPLD data model:

    - :obj:`None` for the `Null kind <https://ipld.io/docs/data-model/kinds/#null-kind>`_
    - :obj:`bool` for the `Boolean kind <https://ipld.io/docs/data-model/kinds/#boolean-kind>`_
    - :obj:`int` for the `Integer kind <https://ipld.io/docs/data-model/kinds/#integer-kind>`_
    - :obj:`float` for the `Float kind <https://ipld.io/docs/data-model/kinds/#float-kind>`_
    - :obj:`str` for the `String kind <https://ipld.io/docs/data-model/kinds/#string-kind>`_
    - :obj:`bytes` for the `Bytes kind <https://ipld.io/docs/data-model/kinds/#bytes-kind>`_
    - :class:`CID` for the `Link kind <https://ipld.io/docs/data-model/kinds/#link-kind>`_

"""

IPLDKind = Union[IPLDScalarKind, List["IPLDKind"], Dict[str, "IPLDKind"]]
r"""
    Python type alias for `kinds <https://ipld.io/docs/data-model/kinds/>`_ in the IPLD data model:

    - :obj:`None` for the `Null kind <https://ipld.io/docs/data-model/kinds/#null-kind>`_
    - :obj:`bool` for the `Boolean kind <https://ipld.io/docs/data-model/kinds/#boolean-kind>`_
    - :obj:`int` for the `Integer kind <https://ipld.io/docs/data-model/kinds/#integer-kind>`_
    - :obj:`float` for the `Float kind <https://ipld.io/docs/data-model/kinds/#float-kind>`_
    - :obj:`str` for the `String kind <https://ipld.io/docs/data-model/kinds/#string-kind>`_
    - :obj:`bytes` for the `Bytes kind <https://ipld.io/docs/data-model/kinds/#bytes-kind>`_
    - :class:`CID` for the `Link kind <https://ipld.io/docs/data-model/kinds/#link-kind>`_
    - :obj:`List` for the `List kind <https://ipld.io/docs/data-model/kinds/#list-kind>`_
    - :obj:`Dict` for the `Map kind <https://ipld.io/docs/data-model/kinds/#map-kind>`_

"""

IPLDObjPathSegment = Union[int, str]
r"""
    An individual segment in a :class:`IPLDObjPath` within a IPLD value (see :obj:`IPLDKind` for the ). A segment can be an :obj:`int` or a :obj:`str`:

    - an :obj:`int` segment is a position, indexing an item in a value of List :obj:`IPLDKind` (a :obj:`List` in Python)
    - an :obj:`str` segment is a key, indexing a value in a value of Map :obj:`IPLDKind` (a :obj:`Dict` in Python)

"""

_IPLDObjPathSegments = Tuple[IPLDObjPathSegment, ...]
r"""
    Short type alias for multiple segments.
"""

class IPLDObjPath(Sequence[IPLDObjPathSegment]):
    r"""
        Path within an object of :obj:`IPLDKind`, as a sequence of :obj:`IPLDObjPathSegment`.
        Paths are immutable and hashable, and a path is a :obj:`Sequence` of the segments that constitute it.
    """

    _instances: ClassVar[MutableMapping[_IPLDObjPathSegments, IPLDObjPath]] = WeakValueDictionary()

    @staticmethod
    def parse(path_str: str) -> IPLDObjPath:
        r"""
            Parses a :class:`IPLDObjPath` from a string representation where segments are separated by `"/"`, such as that returned by
            :meth:`IPLDObjPath.__repr__`.
        """
        if path_str.startswith("IPLDObjPath()"):
            path_str = path_str[6:]
        if not path_str.startswith("/"):
            raise ValueError("Path must start with '/' or 'IPLDObjPath()/'.")
        segs: List[IPLDObjPathSegment] = []
        seg_str_list = path_str[1:].split("/")
        for idx, seg_str in enumerate(seg_str_list):
            if seg_str.startswith("'"):
                if not seg_str.endswith("'"):
                    raise ValueError(f"At segment {idx}: opening single quote without closing single quote.")
                segs.append(seg_str[1:-1])
            elif seg_str.startswith('"'):
                if not seg_str.endswith('"'):
                    raise ValueError(f"At segment {idx}: opening double quote without closing double quote.")
                segs.append(seg_str[1:-1])
            else:
                if not seg_str.isnumeric():
                    raise ValueError(f"At segment {idx}: segment is unquoted and not numeric.")
                segs.append(int(seg_str))
        return IPLDObjPath._new_instance(tuple(segs))

    @staticmethod
    def _new_instance(segments: Tuple[IPLDObjPathSegment, ...]) -> IPLDObjPath:
        r"""
            Returns an instance of :class:`IPLDObjPath` with given segments, without performing any validation.
        """
        instance = IPLDObjPath._instances.get(segments)
        if instance is None:
            instance = object.__new__(IPLDObjPath)
            instance._segments = segments
            IPLDObjPath._instances[segments] = instance
        return instance

    _segments: _IPLDObjPathSegments

    def __new__(cls, *segments: IPLDObjPathSegment) -> IPLDObjPath:
        r""" Constructor for :class:`IPLDObjPath`. """
        validate(segments, _IPLDObjPathSegments)
        return IPLDObjPath._new_instance(segments)

    def access(self, value: IPLDKind) -> IPLDKind:
        r"""
            Accesses the sub-value at this path in the given IPLD value.
            Can be written more expressively as `self >> value`, see :meth:`IPLDObjPath.__rshift__`.
        """
        return _access(self, value)

    def __truediv__(self, other: Union[IPLDObjPathSegment, IPLDObjPath]) -> IPLDObjPath:
        r"""
            The `/` operator can be used to create paths by concatenating segments. Below we use `_` as a suggestive name for an empty path, acting as root:

            >>> _ = IPLDObjPath()
            >>> p = _/2/'red'
            >>> p
            /2/'red'

            Concatenating an existing path with one or more segments returns a new path, extended by the given segments:

            >>> p/3
            /2/'red'/3
            >>> p/0/'blue'
            /2/'red'/0/'blue'

            Concatenating two paths yields a new path, where the end of the first path is treated as the root for the second:

            >>> q = _/0/'blue'
            >>> p/q
            /2/'red'/0/'blue'
        """
        if isinstance(other, (int, str)):
            return IPLDObjPath._new_instance(self._segments+(other,))
        if isinstance(other, IPLDObjPath):
            return IPLDObjPath._new_instance(self._segments+other._segments)
        return NotImplemented

    def __rtruediv__(self, other: Union[IPLDObjPathSegment, IPLDObjPath]) -> IPLDObjPath:
        r"""
            It is possible to prepend a single segment at a time to an existing path using `/` (a new path is returned):

            >>> _ = IPLDObjPath()
            >>> p = _/2/'red'
            >>> 1/p
            /1/2/'red'

            Prepending multiple segments requires brackets (because the `/` operator associates to the left):

            >>> 0/(1/p)
            /0/1/2/'red'
        """
        if isinstance(other, (int, str)):
            return IPLDObjPath._new_instance((other,)+self._segments)
        return NotImplemented

    def __len__(self) -> int:
        return len(self._segments)

    def __iter__(self) -> Iterator[IPLDObjPathSegment]:
        return iter(self._segments)

    @overload
    def __getitem__(self, idx: int) -> IPLDObjPathSegment:
        ...

    @overload
    def __getitem__(self, idx: slice) -> IPLDObjPath:
        ...

    def __getitem__(self, idx: Union[int, slice]) -> Union[IPLDObjPathSegment, IPLDObjPath]:
        if isinstance(idx, int):
            return self._segments[idx]
        return IPLDObjPath._new_instance(self._segments[idx])

    def __le__(self, other: IPLDObjPath) -> bool:
        r"""
            The `<` and `<=` operators can be used to check whether a path is a (strict) sub-path of another path, starting at the same root:

            >>> _ = IPLDObjPath()
            >>> p = _/0/'red'
            >>> q = p/1/2
            >>> p == q
            False
            >>> p <= q
            True
            >>> p < q
            True

        """
        if isinstance(other, IPLDObjPath):
            return len(self) <= len(other) and all(a == b for a, b in zip(self, other))
        return NotImplemented

    def __lt__(self, other: IPLDObjPath) -> bool:
        r""" See :meth:`IPLDObjPath.__le__`. """
        if isinstance(other, IPLDObjPath):
            return len(self) < len(other) and all(a == b for a, b in zip(self, other))
        return NotImplemented

    def __repr__(self) -> str:
        r"""
            .. code-block:: python

                return "/"+"/".join(repr(seg) for seg in self)
        """
        return "/"+"/".join(repr(seg) for seg in self)

    def __rshift__(self, value: IPLDKind) -> IPLDKind:
        r"""
            Accesses the sub-value at this path in the given IPLD value:

            >>> _ = IPLDObjPath()
            >>> _ >> [0, False, {"a": b"hello", "b": "bye"}]
            [0, False, {'a': b'hello', 'b': 'bye'}]
            >>> _/2 >> [0, False, {"a": b"hello", "b": "bye"}]
            {'a': b'hello', 'b': 'bye'}
            >>> _/2/'b' >> [0, False, {"a": b"hello", "b": "bye"}]
            'bye'

            :raises ValueError: if attempting to access a sub-value in a value of :obj:`IPLDScalarKind`
            :raises ValueError: if attempting to access a sub-value indexed by a :obj:`str` segment in a value of list :obj:`IPLDKind` (a Python :obj:`List`)
            :raises ValueError: if attempting to access a sub-value keyed by a :obj:`int` segment in a value of map :obj:`IPLDKind` (a Python :obj:`Dict`)
            :raises IndexError: if attempting to access a sub-value in a value of list kind, where the :obj:`int` segment is not a valid index for the list
            :raises KeyError: if attempting to access a sub-value in a value of map kind, where the :obj:`str` segment is not a valid key for the map
            :raises TypeError: if any of the sub-values along the path is not of IPLD :obj:`IPLDKind` at the top level
        """
        return _access(self, value)


_scalar_kinds = (type(None), bool, int, float, str, bytes, CID)
_recursive_kinds = (list, dict)

def _access(path: IPLDObjPath, value: IPLDKind, idx: int = 0) -> IPLDKind:
    r"""
        Implementation for :func:`IPLDObjPath.access` and :func:`IPLDObjPath.__rshift__`.
    """
    if isinstance(value, _scalar_kinds):
        if len(path) > idx:
            err = f"Error trying to access value at {path[:idx+1]}: value at {path[:idx]} is of scalar kind."
            raise ValueError(err)
        return value
    if isinstance(value, list):
        if idx >= len(path):
            return value
        key = path[idx]
        if not isinstance(key, int):
            err = f"Error trying to access value at {path[:idx+1]}: value at {path[:idx]} is of list kind, but segment {repr(path[idx])} is not integer."
            raise ValueError(err)
        if key not in range(len(value)):
            err = f"Error trying to access value at {path[:idx+1]}: segment {repr(path[idx])} is not a valid index for list at {path[:idx]}."
            raise IndexError(err)
        return _access(path, value[key], idx + 1)
    if isinstance(value, dict):
        if idx >= len(path):
            return value
        key = path[idx]
        if not isinstance(key, str):
            err = f"Error trying to access value at {path[:idx+1]}: value at {path[:idx]} is of map kind, but segment {repr(path[idx])} is not a string."
            raise ValueError(err)
        if key not in value:
            err = f"Error trying to access value at {path[:idx+1]}: segment {repr(path[idx])} is not a valid key for map at {path[:idx]}."
            raise KeyError(err)
        return _access(path, value[key], idx + 1)
    err = f"Error trying to access value at {path[:idx+1]}: value at {path[:idx]} is not of IPLD kind (found type {type(value)})."
    raise TypeError(err)



"""
    Deconding function for DAG-CBOR codec.
"""

from __future__ import annotations # See https://peps.python.org/pep-0563/

from io import BufferedIOBase, BytesIO
import math
import struct
import sys
from typing import Any, Dict, Callable, List, Optional, Sequence, Tuple, Union, cast
import unicodedata
from typing_extensions import Literal, Protocol, TypedDict
from typing_validation import validate

from multiformats import multicodec, CID, varint

from ..ipld import IPLDKind
from ..encoding import _dag_cbor_code
from .err import CBORDecodingError, DAGCBORDecodingError
from . import _err
from ._stream import Stream

__all__ = ("CBORDecodingError", "DAGCBORDecodingError")

class DecodeCallback(Protocol):
    r"""
        Type of optional callbacks for the :func:`decode` function.
    """
    def __call__(self, value: IPLDKind, num_bytes_read: int) -> None:
        ...

class _DecodeOptions(TypedDict, total=False):
    r""" Options passed around to decoding sub-routines. """

    callback: "DecodeCallback"
    r""" An optional callback to be called on each decoded item. """

    normalize_strings: Literal["NFC", "NFKC", "NFD", "NFKD"]
    r""" Optional Unicode normalization to be performed on decoded UTF-8 strings. """

def decode(stream_or_bytes: Union[BufferedIOBase, bytes], *,
           allow_concat: bool = False,
           callback: Optional["DecodeCallback"] = None,
           require_multicodec: bool = False,
           normalize_strings: Literal["NFC", "NFKC", "NFD", "NFKD", None] = None) -> IPLDKind:
    r"""
        Decodes and returns a single data item from the given ``stream_or_bytes``, with the DAG-CBOR codec.

        A simple use for the optional ``callback`` argument is to count the number of bytes read from the stream:

        >>> import dag_cbor
        >>> from io import BytesIO
        >>> class BytesReadCounter:
        ...     _num_bytes_read = 0
        ...     def __call__(self, _, num_bytes_read):
        ...         self._num_bytes_read += num_bytes_read
        ...     def __int__(self):
        ...         return self._num_bytes_read
        ...
        >>> encoded_bytes = b'\xa2aa\x0cabfhello!\x82\x00\x01'
        >>> len(encoded_bytes)
        16
        >>> stream = BytesIO(encoded_bytes)
        >>> bytes_read_cnt = BytesReadCounter()
        >>> dag_cbor.decode(stream, allow_concat=True, callback=bytes_read_cnt)
        {'a': 12, 'b': 'hello!'}
        >>> int(bytes_read_cnt)
        13
        >>> bytes_remaining = stream.read()
        >>> bytes_remaining
        b'\x82\x00\x01'
        >>> len(bytes_remaining)
        3
        >>> dag_cbor.decode(bytes_remaining)
        [0, 1]

        :param stream_or_bytes: the bytes object or bytes stream to decode
        :param allow_concat: whether to allow partial stream decoding (if this is :obj:`False`, a byte stream will always be consumed in its entirety)
        :param callback: optional callback to be invoked as ``callback(item, num_bytes_read)`` every time an item is decoded,
                         where ``num_bytes_read`` is the number of bytes read decoding the item (excluding sub-items, in the case of lists or dictionaries).
        :param require_multicodec: if :obj:`True`, the data being decoded must be prefixed by the multicodec code for ``'dag-cbor'``
                                   (see `multicodec.unwrap <https://multiformats.readthedocs.io/en/latest/api/multiformats.multicodec.html#unwrap>`_).
        :param normalize_strings: whether strings should be normalised after decoding

        :raises CBORDecodingError: while reading the leading byte of a data item head, if no bytes are available
        :raises CBORDecodingError: while reading the argument bytes of a data item head,
                                                   if the expected number of argument bytes is not available
        :raises CBORDecodingError: while decoding the data of a bytestring or string, if the expected number of data bytes is not available
        :raises CBORDecodingError: while decoding the items of a list or a map (keys and values),
                                                   if the expected number of items is not available
        :raises CBORDecodingError: if an invalid utf-8 byte sequence is encountered while attempting to decode a string
        :raises DAGCBORDecodingError: if attempting to decode the special :obj:`float` values ``NaN``, ``Infinity`` and ``-Infinity``
        :raises DAGCBORDecodingError: if the additional info is greater than 27, or different from 27 for major type 7
        :raises DAGCBORDecodingError: if an integer value was not minimally encoded
        :raises DAGCBORDecodingError: if a key of a map is not a string
        :raises DAGCBORDecodingError: if a map has repeated keys
        :raises DAGCBORDecodingError: if map keys are not in canonical order
        :raises DAGCBORDecodingError: if a tag (major type 6) different than 42 (for CID data) is encountered
        :raises DAGCBORDecodingError: if non-bytestring data is found where CID data is expected (tag 42)
        :raises DAGCBORDecodingError: if a simple value (major type 7) different from 20 (False), 21 (True) or 22 (None) is encountered
        :raises DAGCBORDecodingError: if ``require_multicodec`` is set to :obj:`True` and
                                                      the bytes are not prefixed by the ``'dag-cbor'`` multicodec code
        :raises DAGCBORDecodingError: if ``allow_concat`` is set to :obj:`False` and the decoding did not use all available bytes

    """
    validate(stream_or_bytes, Union[BufferedIOBase, bytes])
    validate(allow_concat, bool)
    validate(require_multicodec, bool)
    options: _DecodeOptions = {}
    if callback is not None:
        options["callback"] = callback
    if normalize_strings is not None:
        validate(normalize_strings, Literal["NFC", "NFKC", "NFD", "NFKD"])
        options["normalize_strings"] = normalize_strings
    if isinstance(stream_or_bytes, bytes):
        _stream: BufferedIOBase = BytesIO(stream_or_bytes)
    else:
        _stream = stream_or_bytes
    if require_multicodec:
        code, _, _stream = multicodec.unwrap_raw(_stream)
        stream = Stream(_stream, varint.encode(code))
        if code != _dag_cbor_code:
            raise DAGCBORDecodingError(_err._required_multicodec(stream))
    else:
        stream = Stream(_stream)
    data, _ = _decode_item(stream, options)
    if not allow_concat:
        remaining_bytes = stream.read()
        if len(remaining_bytes) > 0:
            raise DAGCBORDecodingError(_err._multiple_top_level_items(stream))
    return data

def _decode_item(stream: Stream, options: _DecodeOptions) -> Tuple[IPLDKind, int]:
    major_type, arg, num_bytes_read = _decode_head(stream)
    ret: Optional[Tuple[IPLDKind, int]] = None
    assert 0x0 <= major_type <= 0x7, f"Major type must be one of 0x0-0x7, found 0x{major_type:x} instead."
    if isinstance(arg, float):
        # Major type 0x7 (float case):
        assert major_type == 0x7, f"Major type for float must be 0x7, found 0x{major_type:x} instead."
        if math.isnan(arg) or math.isinf(arg):
            raise DAGCBORDecodingError(_err._invalid_float(stream, arg))
        ret = (arg, num_bytes_read)
    elif major_type <= 0x1:
        # Major types 0x0 and 0x1:
        ret = (arg if major_type == 0x0 else -1-arg, num_bytes_read)
    else:
        # Major types 0x2-0x6 and 0x7 (bool/null case):
        value, num_bytes_further_read = _decoders[major_type](stream, arg, options)
        ret = (value, num_bytes_read+num_bytes_further_read)
    if "callback" in options:
        options["callback"](*ret)
    return ret


def _decode_head(stream: Stream) -> Tuple[int, Union[int, float], int]:
    # pylint: disable = too-many-branches
    # read leading byte
    res = stream.read(1)
    if len(res) < 1:
        raise CBORDecodingError(_err._unexpected_eof(stream, "leading byte of data item head", 1, include_prev_snapshot=False))
    leading_byte = res[0]
    major_type = leading_byte >> 5
    additional_info = leading_byte & 0b11111
    # read argument value and return (major_type, arg, num_bytes_read)
    if additional_info < 24:
        # argument value = additional info
        return (major_type, additional_info, 1)
    if additional_info > 27 or (major_type == 0x7 and additional_info != 27):
        raise DAGCBORDecodingError(_err._invalid_additional_info(stream, additional_info, major_type))
    argument_nbytes = 1<<(additional_info-24)
    res = stream.read(argument_nbytes)
    if len(res) < argument_nbytes:
        raise CBORDecodingError(_err._unexpected_eof(stream, f"{argument_nbytes} byte argument of data item head", argument_nbytes))
    if additional_info == 24:
        # 1 byte of unsigned int argument value to follow
        if res[0] < 24:
            raise DAGCBORDecodingError(_err._excessive_int_size(stream, res[0], 1, 0))
        return (major_type, res[0], 2)
    if additional_info == 25:
        # 2 bytes of unsigned int argument value to follow
        arg = struct.unpack(">H", res)[0]
        if arg <= 255:
            raise DAGCBORDecodingError(_err._excessive_int_size(stream, arg, 2, 1))
        return (major_type, arg, 3)
    if additional_info == 26:
        # 4 bytes of unsigned int argument value to follow
        arg = struct.unpack(">L", res)[0]
        if arg <= 65535:
            if arg <= 255:
                raise DAGCBORDecodingError(_err._excessive_int_size(stream, arg, 4, 1))
            raise DAGCBORDecodingError(_err._excessive_int_size(stream, arg, 4, 2))
        return (major_type, arg, 5)
    # necessarily additional_info == 27
    if major_type == 0x7:
        # 8 bytes of float argument value to follow
        return (major_type, struct.unpack(">d", res)[0], 9)
    # 8 bytes of unsigned int argument value to follow
    arg = struct.unpack(">Q", res)[0]
    if arg <= 4294967295:
        if arg <= 255:
            raise DAGCBORDecodingError(_err._excessive_int_size(stream, arg, 8, 1))
        if arg <= 65535:
            raise DAGCBORDecodingError(_err._excessive_int_size(stream, arg, 8, 2))
        raise DAGCBORDecodingError(_err._excessive_int_size(stream, arg, 8, 4))
    return (major_type, arg, 9)

def _decode_bytes(stream: Stream, length: int, options: _DecodeOptions) -> Tuple[bytes, int]:
    res = stream.read(length)
    if len(res) < length:
        raise CBORDecodingError(_err._unexpected_eof(stream, f"{length} bytes of bytestring", length))
    return (res, length)

def _decode_str(stream: Stream, length: int, options: _DecodeOptions) -> Tuple[str, int]:
    res = stream.read(length)
    if len(res) < length:
        raise CBORDecodingError(_err._unexpected_eof(stream, f"{length} bytes of string", length))
    try:
        s = res.decode(encoding="utf-8", errors="strict")
    except UnicodeDecodeError as e:
        raise CBORDecodingError(_err._unicode(stream, length, e.start, e.end, e.reason)) # pylint: disable = raise-missing-from
    if "normalize_strings" in options:
        s = unicodedata.normalize(options["normalize_strings"], s)
    return (s, length)

def _decode_list(stream: Stream, length: int, options: _DecodeOptions) -> Tuple[List[Any], int]:
    list_head_snapshot = stream.curr_snapshot
    l: List[Any] = []
    for idx in range(length):
        try:
            item, _ = _decode_item(stream, options)
            l.append(item)
        except CBORDecodingError as e:
            raise CBORDecodingError(_err._list_item(list_head_snapshot, idx, length, e)) # pylint: disable = raise-missing-from
    return (l, 0)

def _decode_dict_key(stream: Stream, key_idx: int, dict_length: int, options: _DecodeOptions) -> Tuple[str, int, bytes]:
    # pylint: disable = too-many-return-statements, too-many-branches
    major_type, arg, num_bytes_read = _decode_head(stream)
    ret: Optional[Tuple[IPLDKind, int]] = None
    if major_type != 0x3:
        raise DAGCBORDecodingError(_err._dict_key_type(stream, major_type))
    assert not isinstance(arg, float)
    str_length = arg
    str_bytes: bytes = stream.read(str_length)
    if len(str_bytes) < str_length:
        raise CBORDecodingError(_err._unexpected_eof(stream, f"{str_length} bytes of string", str_length))
    try:
        s = str_bytes.decode(encoding="utf-8", errors="strict")
    except UnicodeDecodeError as e:
        raise CBORDecodingError(_err._unicode(stream, str_length, e.start, e.end, e.reason)) # pylint: disable = raise-missing-from
    if "normalize_strings" in options:
        s = unicodedata.normalize(options["normalize_strings"], s)
    ret = (s, num_bytes_read+str_length)
    if "callback" in options:
        options["callback"](*ret)
    return ret+(str_bytes,)

def _decode_dict(stream: Stream, length: int, options: _DecodeOptions) -> Tuple[Dict[str, Any], int]:
    # pylint: disable = too-many-locals
    dict_head_snapshot = stream.curr_snapshot
    d: Dict[str, Any] = {}
    key_bytes_list: List[bytes] = []
    for i in range(length):
        try:
            k, _, k_bytes = _decode_dict_key(stream, i, length, options)
        except CBORDecodingError as e:
            raise CBORDecodingError(_err._dict_item(dict_head_snapshot, "key", i, length, e)) # pylint: disable = raise-missing-from
        if k in d:
            raise DAGCBORDecodingError(_err._duplicate_dict_key(dict_head_snapshot, stream, k, i, length))
        try:
            v, _ = _decode_item(stream, options)
        except CBORDecodingError as e:
            raise CBORDecodingError(_err._dict_item(dict_head_snapshot, "value", i, length, e)) # pylint: disable = raise-missing-from
        d[k] = v
        key_bytes_list.append(k_bytes)
    # check that keys are sorted canonically
    assert len(key_bytes_list) == length
    sorted_key_bytes_list = sorted(key_bytes_list, key=lambda e: (len(e), e))
    for idx0, (kb0, kb1) in enumerate(zip(key_bytes_list, sorted_key_bytes_list)):
        if kb0 != kb1:
            idx1 = key_bytes_list.index(kb1)
            raise DAGCBORDecodingError(_err._dict_key_order(dict_head_snapshot, kb0, idx0, kb1, idx1, length))
    return (d, 0)

def _decode_cid(stream: Stream, arg: int, options: _DecodeOptions) -> Tuple[CID, int]:
    if arg != 42:
        raise DAGCBORDecodingError(_err._invalid_tag(stream, arg))
    cid_head_snapshots = stream.prev_snapshot, stream.curr_snapshot
    try:
        if "callback" in options:
            options = cast(_DecodeOptions, {**options})
            del options["callback"]
        cid_bytes, num_bytes_read = _decode_item(stream, options)
    except CBORDecodingError as e:
        raise CBORDecodingError(_err._cid(cid_head_snapshots, e)) # pylint: disable = raise-missing-from
    if not isinstance(cid_bytes, bytes):
        raise DAGCBORDecodingError(_err._cid_bytes(cid_head_snapshots, stream, cid_bytes))
    if not cid_bytes[0] == 0:
        raise DAGCBORDecodingError(_err._cid_multibase(cid_head_snapshots, stream, cid_bytes))
    return (CID.decode(cid_bytes[1:]), num_bytes_read)

def _decode_bool_none(stream: Stream, arg: int, options: _DecodeOptions) -> Tuple[Optional[bool], int]:
    if arg == 20:
        return (False, 0)
    if arg == 21:
        return (True, 0)
    if arg == 22:
        return (None, 0)
    raise DAGCBORDecodingError(_err._simple_value(stream, arg))

def _decode_dummy(stream: Stream, arg: int, options: _DecodeOptions) -> Tuple[None, int]:
    assert False, f"Major type {arg} does not have an associated decoder."

_decoders: Tuple[Callable[[Stream, int, _DecodeOptions], Tuple[IPLDKind, int]], ...] = (
    _decode_dummy,
    _decode_dummy,
    _decode_bytes,
    _decode_str,
    _decode_list,
    _decode_dict,
    _decode_cid,
    _decode_bool_none
)



r"""
    Errors for the :mod:`dag_cbor.encoding` submodule.
"""

class CBORError(Exception):
    """
        Parent class for all errors due to the CBOR specification.
    """

class DAGCBORError(CBORError):
    """
        Parent class for all errors due to the DAG-CBOR specification.
    """

class CBOREncodingError(CBORError):
    """
        Class for encoding errors due to the CBOR specification.
    """

class DAGCBOREncodingError(CBOREncodingError, DAGCBORError):
    """
        Class for encoding errors due to the DAG-CBOR specification.
    """



r"""
    Utility functions used to produce messages for DAG-CBOR decoding errors.
"""

from __future__ import annotations # See https://peps.python.org/pep-0563/

from typing import List, Optional, Tuple

from .err import CBORDecodingError
from ._stream import StreamSnapshot

_TRUNC_BYTES = 16

def _bytes2hex(bs: bytes) -> str:
    r"""
        Converts bytes to a hex string, showing a truncated string if the number of bytes exceeds 16.
    """
    if len(bs) <= _TRUNC_BYTES:
        return bs.hex()
    return bs[:1].hex()+"..."+bs[-1:].hex() # fixed length 7 < 2*_TRUNC_BYTES

def _decode_error_msg_lines(*snapshots: StreamSnapshot, details: Optional[str] = None,
                            start: int = 0,
                            end: Optional[int] = None,
                            pad_start: int = 0,
                            hl_start: int = 0,
                            hl_len: Optional[int] = None,
                            dots: bool = False,
                            ) -> List[str]:
    r"""
        This utility function takes one or more stream snapshots as input and collates the chunks of read bytes. Let ``bs`` be the bytes read across the
        snapshots (which are assumed to be sequential) and ``pos`` be the position in the stream of the first byte of ``bs`` (he ``start`` and ``end`` arguments
        can be used to focus on a sub-range):

        .. code-block:: python

            bs = b"".join((snapshot.latest_read for snapshot in snapshots))[start:end]
            pos = snapshots[0].latest_read_start+start

        Th utility function returns one or two lines of error message, based on that information:

        1. The first line shows (a selection of) the bytes read and the position of the first byte in the stream
        2. If ``details`` is specified, the second line highlights a selection of bytes from the first line, followed by the given details

        The following optional keyword arguments can be used to customise the selection of read bytes and the highlighting.

        - The ``pad_start`` arguments can be used to specify whitespace padding at the start of the bytes shown, to align them to bytes on other error lines.
        - The ``hl_start`` and ``hl_len`` arguments can be used to specify the start byte and lenght of the range of bytes highlighted in the second line.
        - The ``dots`` argument can be used to specify that three dots '...' should be added after the bytes, to indicate continuation.

    """
    # pylint: disable = too-many-locals, too-many-arguments
    assert snapshots
    bs = b"".join((snapshot.latest_read for snapshot in snapshots))[start:end]
    pos = snapshots[0].latest_read_start+start
    bs_str = _bytes2hex(bs)
    truncated = len(bs_str) != 2*len(bs)
    if not bs_str:
        bs_str = "<EOF>"
        bs_tab = "^"*len(bs_str)
    else:
        if hl_len is None:
            hl_len = len(bs)-hl_start
        else:
            assert 0 <= hl_len <= len(bs)-start
        if truncated and not (hl_len == 1 and (hl_start in {0, len(bs)-1})):
            bs_tab = "^"*len(bs_str)
        else:
            bs_tab = "  "*hl_start+"^^"*hl_len
    bs_str = "  "*pad_start+bs_str
    bs_tab = "  "*pad_start+bs_tab
    bytes_line = f"At byte #{pos}: {bs_str}"
    if truncated:
        last_byte_idx = pos+len(bs)-1
        bytes_line += f" (last byte #{last_byte_idx})"
    if dots:
        bytes_line += "..."
    lines = [bytes_line]
    if details is not None:
        details_line = f"         {' '*len(str(pos))}  {bs_tab} {details}"
        lines.append(details_line)
    return lines

def _decode_error_msg(msg: str, *snapshots: StreamSnapshot, details: Optional[str] = None,
                      start: int = 0,
                      end: Optional[int] = None,
                      hl_start: int = 0,
                      hl_len: Optional[int] = None,
                      dots: bool = False,
                      ) -> str:
    r"""
        Creates a detailed, multi-line error message, starting from a given ``msg`` and taking into account the information from one or more stream snapshots.
        The resulting error message has ``msg`` on the first line, followed by the lines returned by :func:`_decode_error_msg_lines`.
    """
    # pylint: disable = too-many-arguments
    lines = [msg]
    if snapshots:
        lines.extend(_decode_error_msg_lines(*snapshots, details=details,
                                         start=start, end=end, hl_start=hl_start, hl_len=hl_len,
                                         dots=dots))
    return "\n".join(lines)


def _extract_error_cause_lines(e: CBORDecodingError) -> List[str]:
    r""" Extracts lines of error description from a :class:`CBORDecodingError`. """
    lines = str(e).split("\n")
    return [(r"\ " if idx == 0 else "  ")+line for idx, line in enumerate(lines)]


def _cid_error_template(cid_head_snapshots: Tuple[StreamSnapshot, StreamSnapshot], *explanation: str) -> str:
    r""" Template for CID errors. """
    lines = [
        "Error while decoding CID.",
        *_decode_error_msg_lines(*cid_head_snapshots, details="CID tag", dots=True),
        *explanation
    ]
    return "\n".join(lines)



"""
    Python implementation of the `DAG-CBOR codec <https://ipld.io/specs/codecs/dag-cbor/spec/>`_ specification.
"""

from __future__ import annotations # See https://peps.python.org/pep-0563/

__version__ = "0.3.3"

from .ipld import IPLDKind, IPLDScalarKind, IPLDObjPath
from .encoding import encode
from .decoding import decode

# explicit re-exports
__all__ = ["encode", "decode", "IPLDKind", "IPLDScalarKind", "IPLDObjPath"]




r"""
    Autodoc extension dealing with local type references and function signatures.
"""

from __future__ import annotations

from collections.abc import Callable, Mapping
from dataclasses import dataclass
import inspect
import re
from types import FunctionType, ModuleType
from typing import Any, Optional, Union
from typing_extensions import Literal
from sphinx.application import Sphinx

@dataclass(frozen=True)
class ParsedType:
    r""" Dataclass for a parsed type. """
    name: str
    args: Union[None, str, tuple[ParsedType, ...]] = None
    variadic: bool = False

    def crossref(self, globalns: Optional[Mapping[str, Any]] = None) -> str:
        r""" Generates Sphinx cross-reference link for the given type, using local names. """
        # pylint: disable = eval-used
        if globalns is None:
            globalns = {}
        name, args, variadic = self.name, self.args, self.variadic
        role = "obj"
        if name in globalns:
            obj = globalns[name]
            if isinstance(obj, ModuleType):
                role = "mod"
            elif isinstance(obj, property):
                role = "attr"
            elif isinstance(obj, type):
                role = "class"
            elif isinstance(obj, FunctionType):
                role = "func"
        name_crossref = f":{role}:`{name}`"
        if args is None:
            return name_crossref
        if isinstance(args, str):
            _args = eval(f"({args}, )")
            arg_crossrefs = ", ".join(f"``{repr(arg)}``" for arg in _args)
        else:
            arg_crossrefs = ", ".join((arg.crossref(globalns) for arg in args))
        if variadic:
            arg_crossrefs += ", ..."
        return fr"{name_crossref}\ [{arg_crossrefs}]"

def _find_closing_idx(s: str, c_open: str, c_close: str, idx_open: int = 0) -> int:
    r""" Finds the index where a bracketed/quoted range ends. """
    assert len(c_open) == 1, c_open
    assert len(c_close) == 1, c_close
    assert idx_open < len(s), (idx_open, len(s))
    assert s[idx_open] == c_open, (idx_open, s[idx_open])
    lvl = 1
    idx_close: Optional[int] = None
    for idx in range(idx_open+1, len(s)):
        if s[idx] == c_close:
            lvl -= 1
        elif s[idx] == c_open:
            lvl += 1
        if lvl == 0:
            idx_close = idx
            break
    if idx_close is None:
        error_msg = f"Unbalanced opening symbol found while searching for first outermost {c_open}...{c_close} in {repr(s)}."
        raise ValueError(error_msg)
    return idx_close

def _parse_type(annotation: str) -> tuple[Union[ParsedType, Literal["..."]], str]:
    r""" Parses an annotation into the first type appearing in it, together with the unparsed remainder of the annotation. """
    quote_close_idx = None
    if annotation.startswith("'"):
        quote_close_idx = _find_closing_idx(annotation, "'", "'", 0)
    elif annotation.startswith('"'):
        quote_close_idx = _find_closing_idx(annotation, '"', '"', 0)
    if quote_close_idx is not None:
        if quote_close_idx == 1:
            raise ValueError(f"Cannot parse empty forward reference at start of annotation: annotation = {annotation}")
        annotation = annotation[1:quote_close_idx]+annotation[quote_close_idx+1:]
    try:
        b_open: Optional[int] = annotation.index("[")
    except ValueError:
        b_open = None
    try:
        c_idx = annotation.index(",")
    except ValueError:
        c_idx = None
    if b_open is not None and (c_idx is None or b_open < c_idx):
        b_close = _find_closing_idx(annotation, "[", "]", b_open)
        name = annotation[:b_open]
        args_str = annotation[b_open+1:b_close]
        res = annotation[b_close+1:]
        assert name, (name, args_str, res)
        if name.split(".")[-1] == "Literal":
            return ParsedType(name, args_str), res
        args: list[ParsedType] = []
        variadic = False
        while args_str:
            arg, _args_str = _parse_type(args_str)
            if isinstance(arg, ParsedType):
                args.append(arg)
            else:
                assert arg == "...", arg
                variadic = True
                if _args_str:
                    raise ValueError(f"Ellipsis argument encountered in parametric type, but not at the end of args lis: annotation = {annotation}, args_str = {args_str}")
            if not _args_str:
                break
            if not _args_str.startswith(", "):
                raise ValueError(f"Multiple type parameters must be separated by ', ': annotation = {annotation}, args_str = {args_str}, arg = {arg} _args_str = {_args_str}")
            args_str = _args_str[2:]
        return ParsedType(name, tuple(args), variadic), res
    if c_idx is not None:
        name = annotation[:c_idx]
        res = annotation[c_idx:]
        if name == "...":
            raise ValueError(f"Found ellipsis followed by comma: annotation = {annotation}")
        return ParsedType(name), res
    if "]" in annotation:
        raise ValueError(f"Encountered closing bracket ']' without any opening bracket: annotation = {annotation} ")
    name = annotation
    if name == "...":
        return "...", ""
    return ParsedType(name), ""

def parse_type(annotation: str) -> ParsedType:
    r""" Parses an annotation into a type. """
    parsed_type, residual_string = _parse_type(annotation)
    if residual_string:
        raise ValueError(f"Annotation was not entirely consumed by parsing: annotation = {annotation}, parsed_type = {parsed_type}, residual_string = {residual_string}")
    if not isinstance(parsed_type, ParsedType):
        raise ValueError(f"Cannot parse ellipsis on its own: annotation = {annotation}")
    return parsed_type

def sigdoc(fun: FunctionType, lines: list[str]) -> None:
    r"""
        Returns doclines documenting the parameter and return type of the given function
    """
    # pylint: disable = too-many-branches
    doc = "\n".join(lines)
    lines.append("")
    # FIXME: if an :rtype: line already exists, remove it here and re-append it after all param type lines.
    globalns = fun.__globals__
    sig = inspect.signature(fun)
    for p in sig.parameters.values():
        annotation = p.annotation
        if annotation == p.empty:
            continue
        if not isinstance(annotation, str):
            print(f"WARNING! Found non-string annotation: {repr(annotation)}. Did you forget to import annotation from __future__?.")
            annotation = str(annotation)
        t = parse_type(annotation)
        tx = t.crossref(globalns)
        default = p.default if p.default != p.empty else None
        is_args = p.kind == p.VAR_POSITIONAL
        is_kwargs = p.kind == p.VAR_KEYWORD
        if is_args:
            extra_info = "variadic positional"
        elif is_kwargs:
            extra_info = "variadic keyword"
        elif default is not None:
            extra_info = f"default = ``{default}``"
        else:
            extra_info = None
        if extra_info is None:
            line = f":type {p.name}: {tx}"
        else:
            line = f":type {p.name}: {tx}; {extra_info}"
        if f":param {p.name}:" not in doc:
            lines.append(f":param {p.name}:")
        if f":type {p.name}:" not in doc:
            lines.append(line)
    if sig.return_annotation == sig.empty:
        return
    t = parse_type(sig.return_annotation)
    tx = t.crossref(globalns)
    line = f":rtype: {tx}"
    if ":rtype:" not in doc:
        lines.append(line)

def sigdoc_handler(app: Sphinx, what: str, fullname: str, obj: Any, options: Any, lines: list[str]) -> None:
    r"""
        Handler for Sphinx Autodoc's event
        `autodoc-process-docstring<https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#event-autodoc-process-docstring>`_
        which replaces cross-references specified in terms of module globals with their fully qualified version.
    """
    # pylint: disable = too-many-arguments
    if what not in ("function", "method", "property"):
        return
    if what == "property":
        fun: FunctionType = obj.fget
    else:
        fun = obj
    sigdoc(fun, lines)

def simple_crossref_pattern(name: str) -> re.Pattern[str]:
    r"""
        Pattern for simple imports:

        .. code-block :: python

            f":{role}:`{name}`"        # e.g. ":class:`MyClass`"
            f":{role}:`~{name}`"       # e.g. ":class:`~MyClass`"
            f":{role}:`{name}{tail}`"  # e.g. ":attr:`MyClass.my_property.my_subproperty`"
            f":{role}:`~{name}{tail}`" # e.g. ":attr:`~MyClass.my_property.my_subproperty`"

    """
    return re.compile(rf":([a-z]+):`(~)?{name}(\.[\.a-zA-Z0-9_]+)?`")

def simple_crossref_repl(name: str, fullname: str) -> Callable[[re.Match[str]], str]:
    r"""
        Replacement function for the pattern generated by :func:`simple_crossref_pattern`:

        .. code-block :: python

            f":{role}:`~{fullname}`"                    # e.g. ":class:`~mymod.mysubmod.MyClass`"
            f":{role}:`~{fullname}`"                    # e.g. ":class:`~mymod.mysubmod.MyClass`"
            f":{role}:`{name}{tail}<{fullname}{tail}>`" # e.g. ":attr:`MyClass.my_property.my_subproperty<mymod.mysubmod.MyClass.my_property.my_subproperty>`"
            f":{role}:`~{fullname}{tail}`"              # e.g. ":attr:`~mymod.mysubmod.MyClass.my_property.my_subproperty`"

    """
    def repl(match: re.Match[str]) -> str:
        role = match[1]
        short = match[2] is not None
        tail = match[3]
        if tail is None:
            return f":{role}:`~{fullname}`"
        if short:
            return f":{role}:`~{fullname}{tail}`"
        return f":{role}:`{name}{tail}<{fullname}{tail}>`"
    return repl

def labelled_crossref_pattern(name: str) -> re.Pattern[str]:
    r"""
        Pattern for labelled imports:

        .. code-block :: python

            f":{role}:`{label}<{name}>`"       # e.g. ":class:`my class<MyClass>`"
            f":{role}:`{label}<{name}{tail}>`" # e.g. ":attr:`my_property<MyClass.my_property>`"

    """
    return re.compile(rf":([a-z]+):`([\.a-zA-Z0-9_]+)<{name}(\.[\.a-zA-Z0-9_]+)?>`")

def labelled_crossref_repl(name: str, fullname: str) -> Callable[[re.Match[str]], str]:
    r"""
        Replacement function for the pattern generated by :func:`labelled_crossref_pattern`:

        .. code-block :: python

            f":{role}:`{label}<{fullname}>`"       # e.g. ":class:`my class<mymod.mysubmod.MyClass>`"
            f":{role}:`{label}<{fullname}{tail}>`" # e.g. ":attr:`my_property<mymod.mysubmod.MyClass.my_property>`"

    """
    def repl(match: re.Match[str]) -> str:
        role = match[1]
        label = match[2]
        tail = match[3]
        if tail is None:
            return f":{role}:`{label}<{fullname}>`"
        return f":{role}:`{label}<{fullname}{tail}>`"
    return repl

_crossref_subs: list[tuple[Callable[[str], re.Pattern[str]],
                           Callable[[str, str], Callable[[re.Match[str]], str]]]] = [
    (simple_crossref_pattern, simple_crossref_repl),
    (labelled_crossref_pattern, labelled_crossref_repl),
]
r"""
    Substitution patterns and replacement functions for various kinds of cross-reference scenarios.
"""

def _get_module_by_name(modname: str) -> ModuleType:
    r"""
        Gathers a module object by name.
    """
    # pylint: disable = exec-used, eval-used
    exec(f"import {modname.split('.')[0]}")
    mod: ModuleType = eval(modname)
    if not isinstance(mod, ModuleType):
        return None
    return mod

def _get_obj_mod(app: Sphinx, what: str, fullname: str, obj: Any) -> Optional[ModuleType]:
    r"""
        Gathers the containing module for the given ``obj``.
    """
    autodoc_type_aliases = app.config.__dict__.get("autodoc_type_aliases")
    name = fullname.split(".")[-1]
    obj_mod: Optional[ModuleType]
    if autodoc_type_aliases is not None:
        if name in autodoc_type_aliases and fullname == autodoc_type_aliases[name]:
            modname = ".".join(fullname.split(".")[:-1])
            obj_mod = _get_module_by_name(modname)
            return obj_mod
    if what == "module":
        obj_mod = obj
    elif what in ("function", "class", "method"):
        obj_mod = inspect.getmodule(obj)
    elif what == "property":
        obj_mod = inspect.getmodule(obj.fget)
    elif what == "data":
        modname = ".".join(fullname.split(".")[:-1])
        obj_mod = _get_module_by_name(modname)
    elif what == "attribute":
        modname = ".".join(fullname.split(".")[:-2])
        obj_mod = _get_module_by_name(modname)
    else:
        print(f"WARNING! Encountered unexpected value for what = {what} at fullname = {fullname}")
        obj_mod = None
    return obj_mod

def _build_fullname_dict(app: Sphinx, fullname: str, obj_mod: Optional[ModuleType], ) -> dict[str, str]:
    r"""
        Builds a dictionary of substitutions from module global names to their fully qualified names,
        based on :func:`inspect.getmodule` and `autodoc_type_aliases` (if specified in the Sphinx app config).
    """
    autodoc_type_aliases = app.config.__dict__.get("autodoc_type_aliases")
    fullname_dict: dict[str, str] = {}
    if obj_mod is not None:
        globalns = obj_mod.__dict__
        for g_name, g_obj in globalns.items():
            if isinstance(g_obj, (FunctionType, type)):
                g_mod = inspect.getmodule(g_obj)
            elif isinstance(g_obj, ModuleType):
                g_mod = g_obj
            else:
                g_mod = inspect.getmodule(g_obj)
            if g_mod is None or g_mod == obj_mod:
                continue
            if g_name not in g_mod.__dict__:
                continue
            g_modname = g_mod.__name__
            fullname_dict[g_name] = f"{g_modname}.{g_name}"
    if autodoc_type_aliases is not None:
        for a_name, a_fullname in autodoc_type_aliases.items():
            if a_name not in fullname_dict:
                fullname_dict[a_name] = a_fullname
    return fullname_dict

def local_crossref_handler(app: Sphinx, what: str, fullname: str, obj: Any, options: Any, lines: list[str]) -> None:
    r"""
        Handler for Sphinx Autodoc's event
        `autodoc-process-docstring<https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#event-autodoc-process-docstring>`_
        which replaces cross-references specified in terms of module globals with their fully qualified version.
    """
    # pylint: disable = too-many-arguments, too-many-locals
    obj_mod = _get_obj_mod(app, what, fullname, obj)
    fullname_dict = _build_fullname_dict(app, fullname, obj_mod)
    for sub_name, sub_fullname in fullname_dict.items():
        for idx, line in enumerate(lines):
            for pattern_fun, repl_fun in _crossref_subs:
                pattern = pattern_fun(sub_name)
                repl = repl_fun(sub_name, sub_fullname)
                line = re.sub(pattern, repl, line)
            lines[idx] = line

def setup(app: Sphinx) -> None:
    r"""
        Registers handlers for Sphinx Autodoc's event
        `autodoc-process-docstring<https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#event-autodoc-process-docstring>`_
    """
    app.connect("autodoc-process-docstring", sigdoc_handler)
    app.connect("autodoc-process-docstring", local_crossref_handler)



"""
    Tests on the specifics of DAG-CBOR Link encoding.
"""

# pylint: disable = global-statement

import cbor2

from dag_cbor import encode, decode
from dag_cbor.decoding.err import DAGCBORDecodingError
from dag_cbor.random import rand_cid, options

import pytest

nsamples = 1000

def test_decoding_requires_multibase_prefix() -> None:
    """
        Checks that the decoder fails if the identity multibase prefix (0x00) is not
        present in the DAG-CBOR representation.
    """
    test_data = rand_cid(nsamples)
    for i, x in enumerate(test_data):
        with pytest.raises(DAGCBORDecodingError):
            decode(cbor2.dumps(cbor2.CBORTag(42, bytes(x))))

def test_encoding_produces_multibase_prefix() -> None:
    """
        Checks that the encoder includes the identity multibase prefix (0x00) in the
        DAG-CBOR representation.
    """
    test_data = rand_cid(nsamples)
    for i, x in enumerate(test_data):
        rtdata = cbor2.loads(encode(x))
        assert rtdata.value[0] == 0
